% MATLAB script that produces the think-aloud-related figure and analyses
close all;close all;clc
addpath('..\Functions')
t=readtable('Transcripts_real.xlsx','PreserveVariableNames',true); % read transcripts
lines_to_exclude=find(~isnan(table2array(t(:,4))));
str=t(:,2); % % all rows with phrases
str2=table2cell(str)'; % convert to cell format
p = [".","?","!",",",";",":","...","…","[","]"];
str2=replace(str2,p," "); % remove punctuation
str2=strip(str2); % remove leading and trailing whitespaces
Times=table2array(t(:,3)); % elapsed time per phrase
Participant=NaN(size(Times)); % initialize vector with participant numbers

Starttimes=[52.52 % start times of participants (i.e., elapsed time in the Tobii video that eye-gaze first fell on the painting)
    49.12
    57.64
    56.6
    47.8
    58.24
    40.72
    31.72
    46.8
    41.68
    51.44
    39.32
    58.48
    33.24
    48.32
    41.16
    62.72
    46.64
    161.24
    107.52
    257.96];

for i=1:length(Times)
    if i==1 % first row
        Participant(i)=1;
    elseif Times(i)>Times(i-1) % same participant
        Participant(i)=Participant(i-1);
    else % new participant
        Participant(i)=Participant(i-1)+1;
    end % elapsed time since start of session for participant i
    Times(i)=Times(i)-Starttimes(Participant(i));
end
[strWords,strWordst,strWordsp]=deal([]);
[starttime_of_statement,endtime_of_statement]=deal(NaN(size(str2,1),1));
for i=1:length(str2) % loop over all phrases
    starttime_of_statement(i)=Times(i);
    if i==length(str2) % for the last phrase of the last participant, the end time is 300
        endtime_of_statement(i)=300;
    else
        endtime_of_statement(i)=Times(i+1); % the end time of a statement is the start time of the next statement
        if endtime_of_statement(i)<starttime_of_statement(i) % if the end time of a statement is smaller than the start time of the same statement, then i is the end of the last statement of this participant
            endtime_of_statement(i)=300;
        end
    end
    if Participant(i)==16 && endtime_of_statement(i)==300 % manual correction for participant 16; audio ended early due to a flat battery of the Tobii
        endtime_of_statement(i)=243.770-Starttimes(16);
    end

    words_of_statement=split(str2(i));
    words_of_statement_times=linspace(starttime_of_statement(i),endtime_of_statement(i),length(words_of_statement));
    
    if ~ismember(i,lines_to_exclude) % only append words if not from a to-be-excluded phrase (reflection of lamps)
        strWords = [strWords; words_of_statement]; %#ok<*AGROW> % append all separate words
        strWordst = [strWordst; words_of_statement_times']; % append times per separate word
        strWordsp = [strWordsp; Participant(i)*ones(size(words_of_statement_times'));]; % append participant number per separate word
    end
end
strWords = lower(strWords); % convert to lower case characters
%%
Strings={"nachtwacht" % 1. Night Watch
    ["meisje" "meisjes" "dame" "meid" "vrouw" "engel" "engelachtige" "prinses" "dochter"]  % 2. Girl
    ["hond" "hondje" "kat"] % 3. Dog
    ["kip" "duif" "haan" "vogel" "trofee"] % 4. Chicken
    ["prachtig" "schitterend" "indrukwekkend" "bijzonder" "mooi" "mooie" "fascinerend" "enig" "geweldig" "prachtige" "gaaf" "fantastisch" "fantastische" "wauw" "genieten"] % 5. Admiration
    ["details" "detail" "textuur" "gedetailleerd" "penseelstreken" "kwaststreken" "brush" "scherp" "realistisch" "foto" "scherper" "realistische" "gedetailleerdheid" "detailrijk"] % 6. Detail
    ["trommel" "trommelaar" "trommels" "trom" "drummer" "muziekinstrument" "instrumenten" "trommeleer" "trommelen" "trommelier" "muzikant" "getrommel"] % 7. Drum
    ["mannen" "mensen" "man" "figuren" "personen" "karakters" "personage" "personages" "soldaten" "wachter" "wachters" "nachtwachters" "mannetjes" "meneer"] % 8. People
    ["hoed" "helm" "helmen"  "kleding" "kleren" "hand" "jas" "jasje" "jassen" "jurkje" "mouw" "mouwtje" "stof" "stoffen" "kaftan" ... 
     "sjaal" "sjerp" "sjaals" "sjaal-achtig" "sash" "vest" "pak" "pakje" "cave" "cape" "laarzen" "schoenen" "enkelschoentjes" "hakjes" "riem" "band" "handschoen" ...
     "kostuum" "kostuums"  "pofbroek" "kledij" "outfit"  "gewaden" "gewaad" "tenue" "tenuetje" "helmpje" ...
     "hals-sjerp" "trekzak" "sol" "hoofdbedekking" "legerhelmen" ...
     "pluimen"  "vogelveer" "veer" "lauwerkrans" "kranshals" "veren" "pluim" "kraag" "kragen" ...
     "stofuitdrukking" "borduursel" "versiering" "versieringen" "plooitjes"  "randen" "randjes" "vouwen" "knoppen" "het kant" "goudstiksel" "mooiigheden" "sierrandjes"  "stiksels" ] % 9. Clothing
    ["middelste" "hoofdpersoon" "hoofdfiguur" "belangrijk" "belangrijke" "belangrijkste" "centrale" "voorgrond" "vooraan" "voorkant" "voorste" "Banninck" "generaal" "Ruyter" "edellieden"...
    "luitenant" "handlanger" "hoofdfiguren" "rechterman" "kooplieden" "koopman" "koningen" "koning" "leider" "bondgenoot" "spreker" "baas" "adel"] % 10. Central figures
    ["wit" "witte" "lichter" "contrasterende" "contrasterende" "contrasten" "uitgelicht" "opgelicht" "opgelichte"] % 11. Contrast
    ["zwart" "zwarte" "donker" "donkere" "duisternis" "somber" "schaduw" "schaduwwerking" "schaduweffecten" "schim" "schimmen" "schemerigere" "schaduwachtige" "schaduwen"] % 12. Darkness
    ["licht" "lichtval" "lichtinval" "lichteffecten" "lichtstraal" "verlicht"  "belicht" "schijnt" "belichten" "lichtbron" "geflitst" "flitser"] % 13. Lighting
    ["kleur" "kleuren" "kleurrijk" "rode" "rood" "blauw" "blauwe" "blauwige" "bruin" "bruinigs" "groen" "groens" "groenige" "blauwiger" "grijs" "geel" "gele" "goud" "gouden" "lichtgeel"] % 14. Colors
    ["wapens" "wapen" "geweer" "geweren" "speer" "speren" "zwaard" "zwaarden" "zwaardachtig" "lans" "lansen" "pistool" "musket" "harpoen" "stokken" "pallen" "balken" "vlag" "vlaggen" "wapenuitrusting"...  
     "pijl" "pijlen" "speerpunten" "vaandel" "stok" "lap" "speer-ding"] % 15. Weapons
    ["gezichten" "gezicht" "koppen" "hoofd" "hoofden" "snor" "oog" "ogen" "uitdrukking" "uitdrukkingen" "gezichtsuitdrukkingen" "sik" "baard" "bebaarde"] % 16. Faces
    ["oorlog" "strijd" "missie" "aanval" "verdedigen" "overwinning" "jacht" "schutterij" "bewaking" "de wacht" "winnen" "overwinnen" "gevecht" "vechten" "aanvallen" "belegerd" "vechtershouding" "beschermen" "vijand" "militair" "slag" "gewonnen" "uittrekken" "geschoten" "opstand" "opporren" "revolutie"]}; % 17. Mission

StringCat={'Night Watch' 'Girl' 'Dog' 'Chicken' 'Admiration' ... % Names of the 17 themes
           'Detail' 'Drum' 'People' 'Clothing' ...
           'Central figures' 'Contrast' 'Darkness' 'Lighting' 'Color' ... 
           'Weapons' 'Faces' 'Mission'};

STATS=NaN(length(Strings),21,3);
for s=1:length(Strings) % loop over 17 themes
    for i=1:21 % loop over participants
        WordsofPerson=find(strWordsp==i); % indexes of participant i
        WordsofPersonIndex=WordsofPerson(contains(strWords(WordsofPerson),Strings{s},'IgnoreCase',true)); % indexes of words of participant i that belong to theme number s
        STATS(s,i,1)=length(WordsofPersonIndex); % number of times the theme is mentioned for participant i 
        if ~isempty(WordsofPersonIndex) % if the theme is mentioned at least once for participant i
            STATS(s,i,2)=mean(strWordst(WordsofPersonIndex)); % mean elapsed time that the theme is addressed
        end
        if s==1
            STATS(1,i,3)=length(WordsofPerson); % total number of words per person
        end 
    end
end

%%
d=squeeze(mean(STATS,2,'omitnan')); % compute average of participants for the five variables (17 x 5 matrix)
[~,b]=sort(d,'descend'); % sort the STATS in descening order (b is 17 x 3 matrix)
b=b(:,1); % sort on the mean number of spoken words per participant per theme

nw1=STATS(b,:,1); % number of spoken words per participant per theme (17 x 21 matrix)
nw1t=STATS(1,:,3); % number of spoken words per participant (1 x 21 vector)
nw1_mt=STATS(b,:,2); % mean time the word is spoken per participant per theme (17 x 21 matrix)

run('Thinkaloud_replica_V2.m') % obtain the data for the Replica

nw2=STATS(b,:,1); % % number of spoken words per participant per theme (17 x 27 matrix)
nw2t=STATS(1,:,3); % number of spoken words per participant (1 x 27 vector)
nw2_mt=STATS(b,:,2); % mean time the word is spoken per participant per theme (17 x 27 matrix)

v1=mean(nw1,2);
v2=mean(nw2,2);
disp('Correlation coefficient of the mean number of words per theme, Real vs. Replica')
disp(round(corr(v1,v2),2))

% Figure 15. Left: Scatter plot of the mean number of spoken words per participant per theme for the experiment with the replica Night Watch versus the real Night Watch. The numbers of the themes are sorted according to the mean number of spoken words in the real painting in descending order. 
colors=distinguishable_colors(length(v1));
figure;
subplot(1,2,1)
plot([0 9],[0 9],'k');
hold on;grid on;box on
clear scatter1 StringCat2

for i=1:length(v1)
    scatter1(i)=scatter(v1(i),v2(i),700,'markerfacecolor',colors(i,:),'markeredgecolor',colors(i,:));hold on;grid on %#ok<SAGROW> 
    scatter1(i).MarkerFaceAlpha = .3; %#ok<SAGROW> 
    text(v1(i),v2(i),num2str(i),'horizontalalignment','center')
    StringCat2{i}=[num2str(i) '. ' char(StringCat(b(i)))]; %#ok<SAGROW> 
end
l1=legend(scatter1,StringCat2);
set(l1,'pos',[0.06 0.51 0.14 0.48])
set(gca,'pos',[0.05 0.09 0.42 0.89])
set(gca,'xlim',[0 9],'ylim',[0 9])
xlabel('Mean number of spoken words per participant (Real painting)')
ylabel('Mean number of spoken words per participant (Replica)')
h=findobj('FontName','Helvetica');set(h,'FontSize',20,'Fontname','Arial')

disp('Number of spoken words Real (all words), Mean SD')
disp([round(mean(nw1t)) round(std(nw1t))])
disp('Number of spoken words Replica (all words), Mean SD')
disp([round(mean(nw2t)) round(std(nw2t))])
[~,p,~,stats]=ttest2(nw1t,nw2t,'vartype','unequal');
disp('Number of spoken words Real vs. Replica (all words), df t p')
disp([round(stats.df,1) round(stats.tstat,2) round(p,3)])

disp('Number of spoken words Real (thematic words), Mean SD')
disp([round(mean(sum(nw1)),1) round(std(sum(nw1)),1)])
disp('Number of spoken words Replica (thematic words), Mean SD')
disp([round(mean(sum(nw2)),1) round(std(sum(nw2)),1)])
[~,p,~,stats]=ttest2(sum(nw1),sum(nw2),'vartype','unequal');
disp('Number of spoken words Real vs. Replica (thematic words), df t p')
disp([round(stats.df,1) round(stats.tstat,2) round(p,3)])

disp('Number of spoken words Real vs. Replica (all 17 themes separately), p df t')
[~,p,~,stats]=ttest2(nw1',nw2','vartype','unequal');
disp([(1:length(p))' round(stats.df',1) round(stats.tstat',2) round(p',3)])

disp('Number of spoken words Real (admiration theme), Mean SD')
disp([round(mean(nw1(b==5,:)),2) round(std(nw1(b==5,:)),2)])
disp('Number of spoken words Replica (admiration theme), Mean SD')
disp([round(mean(nw2(b==5,:)),2) round(std(nw2(b==5,:)),2)])
[~,p,~,stats]=ttest2(nw1',nw2','vartype','unequal');
disp('Number of spoken words Real vs. Replica (admiration theme), df t p')
disp([round(stats.df(b==5),1) round(stats.tstat(b==5),2) round(p(b==5),4)])

% Figure 15. Right: Mean elapsed time of spoken words for the replica Night Watch versus the real Night Watch. The diagonal line is the line of unity.
v1=mean(nw1_mt,2,'omitnan');
v2=mean(nw2_mt,2,'omitnan');
subplot(1,2,2)
plot([0 200],[0 200],'k');
hold on;grid on;box on

clear scatter2
for i=1:length(v1)
    scatter2(i)=scatter(v1(i),v2(i),700,'markerfacecolor',colors(i,:),'markeredgecolor',colors(i,:));hold on;grid on %#ok<SAGROW> 
    scatter2(i).MarkerFaceAlpha = .3; %#ok<SAGROW> 
    text(v1(i),v2(i),num2str(i),'horizontalalignment','center')
end
l2=legend(scatter2,StringCat2,'location','eastoutside');
set(l2,'pos',[0.57 0.51 0.14 0.48])
set(gca,'pos',[0.56 0.09 0.42 0.89])
set(gca,'xtick',0:25:200,'ytick',0:25:200)
xlabel('Mean elapsed time of spoken words (Real painting)')
ylabel('Mean elapsed time of spoken words (Replica)')
h=findobj('FontName','Helvetica');set(h,'FontSize',20,'Fontname','Arial')

filename='Figure15b.png';
set(gcf,'WindowState','maximized');print(gcf,filename,'-dpng','-r600');