% MATLAB script that reads think-aloud data for the experiment with the replica Night Watch. This script does not run independently but is called by Thinkaloud_real.m

t=readtable('Transcripts_replica.xlsx','PreserveVariableNames',true); % read transcripts
str=t(:,2); % all rows with phrases
str2=table2cell(str)'; % convert to cell format
p = [".","?","!",",",";",":","...","…","[","]"];
str2=replace(str2,p," "); % remove punctuation
str2=strip(str2); % remove leading and trailing whitespaces
Times=table2array(t(:,3)); % elapsed time per phrase
Participant=NaN(size(Times)); % initialize vector with participant numbers

Starttimes=[10.24 % start times of participants (i.e., elapsed time in the Tobii video that eye-gaze first fell on the painting)
    2.88
    0.04
    23.68
    18.6
    20.48
    19.56
    8.48
    8.72
    7.96
    20.4
    23.92
    19.32
    18
    17.08
    18.8
    19.4
    25.84
    14.88
    8.72
    1.68
    13.88
    29.28
    20.32
    4.8
    16.96
    23.2];

for i=1:length(Times)
    if i==1 % first row
        Participant(i)=1;
    elseif Times(i)>Times(i-1)
        Participant(i)=Participant(i-1); % same participant
    else
        Participant(i)=Participant(i-1)+1; % new participant
    end
    Times(i)=Times(i)-Starttimes(Participant(i)); % elapsed time since start of session for participant i
end

[strWords,strWordst,strWordsp]=deal([]);
[starttime_of_statement,endtime_of_statement]=deal(NaN(size(str2,1),1));
for i=1:length(str2) % loop over all phrases
    starttime_of_statement(i)=Times(i);
    if i==length(str2) % for the last phrase of the last participant, the end time is 300
        endtime_of_statement(i)=300;
    else
        endtime_of_statement(i)=Times(i+1); % the end time of a statement is the start time of the next statement
        if endtime_of_statement(i)<starttime_of_statement(i) % if the end time of a statement is smaller than the start time of the same statement, then i is the end of the last statement of this participant
            endtime_of_statement(i)=300;
        end
    end

    words_of_statement=split(str2(i),[" ",char(160)]);
    words_of_statement_times=linspace(starttime_of_statement(i),endtime_of_statement(i),length(words_of_statement));
    strWords = [strWords; words_of_statement]; %#ok<*AGROW> % append all separate words
    strWordst = [strWordst; words_of_statement_times']; % append times per separate word
    strWordsp = [strWordsp; Participant(i)*ones(size(words_of_statement_times'));]; % append participant number per separate word
end

iv=find(strcmp(strWords,'')); % find emtpy strongs and remove
strWords(iv)=[];strWordst(iv)=[];strWordsp(iv)=[];

strWords = lower(strWords); % convert to lower case characters

STATS=NaN(length(Strings),27,3);

for s=1:length(Strings) % loop over categories
    for i=1:27 % loop over participants
        WordsofPerson=find(strWordsp==i); % indexes of participant i
        WordsofPersonIndex=WordsofPerson(contains(strWords(WordsofPerson),Strings{s},'IgnoreCase',true)); % indexes of words of participant that belong to theme number s
        STATS(s,i,1)=length(WordsofPersonIndex); % number of times the theme is mentioned for participant i 
        if ~isempty(WordsofPersonIndex) % if the theme is mentioned at least once for participant i
            STATS(s,i,2)=mean(strWordst(WordsofPersonIndex)); % mean elapsed time that the theme is addressed
        end
        if s==1
            STATS(1,i,3)=length(WordsofPerson); % total number of words per person
        end 
    end
end
