% MATLAB script that produces all eye-tracking-related figures and analyses for the experiment with the real Night Watch
clear all;close all;clc %#ok<CLALL>
close all
addpath('..\Functions')

% [~,mapping]=xlsread('..\mapping.xlsx','A2:G22'); % load info about the participants
original=imread('2444px-The_Night_Watch_-_HD.jpg'); % 2444 pixels horizontal, 2048 pixels vertical
original_grayscale=rgb2gray(original);

mycolormap=colormap("hsv");
ft=customcolormap([0 1],[mycolormap(52,:);0.5 0.5 0.5]);
mycolormap(1:52,:)=ft(1:5:end,:);

readraw=0; % if 1, then read Excel files with Tobii data (this will take ~30 minutes and will save a MAT file X.mat)
process_reference_images=0; % if 1, then the reference images from Tobii will be mapped to the frontal image of the painting. Regression coefficients will be stored
match_to_reference_images=0; % if 1, then the Tobii image segment being gazed at is matched with the reference images. The XY coordinates are stored in CData6...mat
plot_for_participant=0; % make plot for a specific frame of a participant (useful for debugging/checking)
video_separate=0; % create video with gaze markers for a single participant on the frontal image of the painting
video_combined=0; % create video with gaze markers of all participants on the frontal image of the painting
recreate2dimage=0; % recreate frontal image of the painting from reference images and save into mat-file

window_size=[300 250 200 150 100 75 50]; % templates of different sizes cropped around the gaze coordinate

for i=1:18 % load the reference images
    fixed_rgb=imread([num2str(i) '.jpg']); % load image
    FIXED_RGB(i,:,:,:)=fixed_rgb; %#ok
    fixed=rgb2gray(fixed_rgb);
    FIXED(i,:,:)=fixed; %#ok
end
%% This block reads Tobii data and saves it in a MAT file X.mat
if readraw==1 % read Excel files with Tobii data
    X=NaN(21,200000,52);
    for p=1:21 % loop over participants
        disp(['Read Tobii data, participant number = ' num2str(p)])
        [x,xt]=xlsread(['Participant' num2str(p) '.xlsx'],'A2:AX300000'); %#ok<XLSRD> 
        x(:,51:52)=[strcmp(xt(:,1),'Accelerometer')+2*strcmp(xt(:,1),'Eye Tracker')+3*strcmp(xt(:,1),'Gyroscope') strcmp(xt(:,35),'fullstream.mp4')];
        X(p,1:size(x,1),1:size(x,2))=x;
    end
    lastindex=find(nansum(X(:,:,1))==0,2,'first'); %#ok<NANSUM>
    X=X(:,1:lastindex(2)-1,:); % cut off empty part of the matrix
    save('X.mat','X') % save MAT file with eye-tracking data
else
    load X.mat
end
%% This block reads the references images from Tobii and matches them to the large frontal image of the painting
%% Regression coefficients are saved. These regression coefficients enable
%% the mapping from the reference images to the frontal image of the painting
%% The regression coefficients are saved in a MAT file RegressionCoefficients
if process_reference_images==1 % read Excel files with Tobii data
    RegressionCoefficients=NaN(18,2,7);
    for i=1:18 % loop over reference images
        disp(['Read and match the references images to the large frontal image of the painting, Reference image number = ' num2str(i)])
        points1 = detectORBFeatures(squeeze(FIXED(i,:,:)),'numlevels',14,'scalefactor',1.2); % detect ORB (i.e., corner) points from reference image
        points2 = detectORBFeatures(original_grayscale,'numlevels',14,'scalefactor',1.2); % detect ORB (i.e., corner) points from frontal image
        [features1,valid_points1] = extractFeatures(squeeze(FIXED(i,:,:)),points1); % extract feature vectors (i.e., descriptors) and their locations from reference image
        [features2,valid_points2] = extractFeatures(original_grayscale,points2); % extract feature vectors (i.e., descriptors) and their locations from frontal image
        indexPairs = matchFeatures(features1,features2); % find matching features between reference and frontal image
        matchedPoints1 = valid_points1(indexPairs(:,1),:); % matched points on reference image
        matchedPoints2 = valid_points2(indexPairs(:,2),:); % matched points on frontal image
        TobiiPoints=matchedPoints1.Location; % location of matched points on reference image
        OriginalPoints=matchedPoints2.Location; % location of matched points on frontal image
        
        switch i
            case 1
                TobiiPoints_corners=[209 772]; % point above the left shoulder of the powder boy
                OriginalPoints_corners=[242 1407]; % point above the left shoulder of the powder boy
                TobiiPoints_corners(2,:)=[1747 193]; % tip of pike at the right (held by Walich Schellingwou)
                OriginalPoints_corners(2,:)=[2268 643]; % tip of pike at the right (held by Walich Schellingwou)
                TobiiPoints_corners(3,:)=[1820 925]; % point at the bottom ring of the drum
                OriginalPoints_corners(3,:)=[2294 1600]; % point at the bottom ring of the drum
                TobiiPoints_corners(4,:)=[156 375]; % nose of figure at the left (Reijnier Engelen)
                OriginalPoints_corners(4,:)=[151 900]; % nose of figure at the left (Reijnier Engelen)
                TobiiPoints_corners(5,:)=[465 45]; % midpoint of flag stick
                OriginalPoints_corners(5,:)=[540 465]; % midpoint of flag stick
                TobiiPoints_corners(6,:)=[1842 155]; % hole at the left edge of the right side of the lining canvas
                OriginalPoints_corners(6,:)=[2395 591]; % hole at the left edge of the right side of the lining canvas
                TobiiPoints_corners(7,:)=[1850 55]; % hole at the rightmost edge of the lining canvas
                OriginalPoints_corners(7,:)=[2414 451]; % hole at the rightmost edge of the lining canvas
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 2
                TobiiPoints_corners=[1208 160]; % right top of the canvas
                OriginalPoints_corners=[2419 26]; % right top of the canvas
                TobiiPoints_corners(2,:)=[29 903]; % left eye of the girl
                OriginalPoints_corners(2,:)=[813 1251]; % left eye of the girl
                TobiiPoints_corners(3,:)=[1267 1057]; % point at the bottom ring of the drum
                OriginalPoints_corners(3,:)=[2340 1570]; % point at the bottom ring of the drum
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 3
                TobiiPoints_corners=[595 126]; % left top of the canvas
                OriginalPoints_corners=[31 33]; % left top of the canvas
                TobiiPoints_corners(2,:)=[1391 10]; % hole at the bottom edge of the top side of the lining canvas
                OriginalPoints_corners(2,:)=[1345 50]; % hole at the bottom edge of the top side of the lining canvas
                TobiiPoints_corners(3,:)=[532 750]; % hole at the right edge of the left side of the lining canvas
                OriginalPoints_corners(3,:)=[47 1114]; % hole at the right edge of the left side of the lining canvas
                TobiiPoints_corners(4,:)=[862 305]; % midpoint of flag stick
                OriginalPoints_corners(4,:)=[541 465]; % midpoint of flag stick
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 4
                TobiiPoints_corners=[628 880]; % left bottom of the painting
                OriginalPoints_corners=[57 1995]; % left bottom of the painting
                TobiiPoints_corners(2,:)=[1780 104]; % nose of pikeman (Walich Schellingwou)
                OriginalPoints_corners(2,:)=[1501 842]; % nose of pikeman (Walich Schellingwou)
                TobiiPoints_corners(3,:)=[1240 996]; % point of the right shoe of the central figure in black (Banninck Cocq)
                OriginalPoints_corners(3,:)=[1016 1959]; % point of the right shoe of the central figure in black (Banninck Cocq)
                TobiiPoints_corners(4,:)=[646 327]; % hole at the right edge of the left side of the lining canvas
                OriginalPoints_corners(4,:)=[47 1114]; % hole at the right edge of the left side of the lining canvas
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 5
                TobiiPoints_corners=[1644 904]; % right bottom of the canvas
                OriginalPoints_corners=[2413 2025]; % right bottom of the canvas
                TobiiPoints_corners(2,:)=[1410 11]; % % tip of pike at the right (held by Walich Schellingwou)
                OriginalPoints_corners(2,:)=[2268 643]; % tip of pike at the right (held by Walich Schellingwou)
                TobiiPoints_corners(3,:)=[64 1]; % midpoint of flag stick
                OriginalPoints_corners(3,:)=[540 480]; % midpoint of flag stick
                TobiiPoints_corners(4,:)=[1488 318]; % left eye of the drummer (Jacob Jorisz)
                OriginalPoints_corners(4,:)=[2310 1108]; % left eye of the drummer (Jacob Jorisz)
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
                invalidindex=find(TobiiPoints(:,1)==826 & TobiiPoints(:,2)==843); % remove invalid marker on the trolley
                TobiiPoints(invalidindex,:)=[];
                OriginalPoints(invalidindex,:)=[];
            case 6
                TobiiPoints_corners=[1210 161]; % right top of the canvas
                OriginalPoints_corners=[2419 26]; % right top of the canvas
                TobiiPoints_corners(2,:)=[35 905]; % left eye of the girl
                OriginalPoints_corners(2,:)=[813 1251]; % left eye of the girl
                TobiiPoints_corners(3,:)=[1236 985]; % hole on drum
                OriginalPoints_corners(3,:)=[2283 1462]; % hole on drum
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 7
                TobiiPoints_corners=[784 185]; % left top of the canvas
                OriginalPoints_corners=[31 33]; % left top of the canvas
                TobiiPoints_corners(2,:)=[719 642]; % nose of figure at the left (Reijnier Engelen)
                OriginalPoints_corners(2,:)=[151 900]; % nose of figure at the left (Reijnier Engelen)
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 8
                TobiiPoints_corners=[504 767]; % left bottom of the painting
                OriginalPoints_corners=[57 1995]; % left bottom of the painting
                TobiiPoints_corners(2,:)=[566 32]; % nose of figure at the left (Reijnier Engelen)
                OriginalPoints_corners(2,:)=[151 900]; % nose of figure at the left (Reijnier Engelen)
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 9
                TobiiPoints_corners=[1242 622]; % right bottom of the canvas
                OriginalPoints_corners=[2413 2025]; % right bottom of the canvas
                TobiiPoints_corners(2,:)=[48 579]; % edge of the butt of the musket (held by Jan van der Heede)
                OriginalPoints_corners(2,:)=[804 1603]; % edge of the butt of the musket (held by Jan van der Heede)
                TobiiPoints_corners(3,:)=[1195 60]; % left eye of the drummer (Jacob Jorisz)
                OriginalPoints_corners(3,:)=[2310 1107]; % left eye of the drummer (Jacob Jorisz)
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 10
                TobiiPoints_corners=[233 807]; % point above the left shoulder of the powder boy
                OriginalPoints_corners=[242 1407]; % point above the left shoulder of the powder boy
                TobiiPoints_corners(2,:)=[1709 225]; % tip of pike at the right (held by Walich Schellingwou)
                OriginalPoints_corners(2,:)=[2268 643]; % tip of pike at the right (held by Walich Schellingwou)
                TobiiPoints_corners(3,:)=[1831 893]; % point at the bottom ring of the drum
                OriginalPoints_corners(3,:)=[2294 1600]; % point at the bottom ring of the drum
                TobiiPoints_corners(4,:)=[179 301]; % nose of figure at the left (Reijnier Engelen)
                OriginalPoints_corners(4,:)=[151 900]; % nose of figure at the left (Reijnier Engelen)
                TobiiPoints_corners(5,:)=[522 62]; % midpoint of flag stick
                OriginalPoints_corners(5,:)=[540 465]; % midpoint of flag stick
                TobiiPoints_corners(6,:)=[1781 199]; % hole at the left edge of the right side of the lining canvas
                OriginalPoints_corners(6,:)=[2395 591]; % hole at the left edge of the right side of the lining canvas
                TobiiPoints_corners(7,:)=[1780 116]; % hole at the rightmost edge of the lining canvas
                OriginalPoints_corners(7,:)=[2414 451]; % hole at the rightmost edge of the lining canvas
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 11 % images 11-18 are created for measurement day 2
                TobiiPoints_corners=[1280 130]; % right top of the canvas
                OriginalPoints_corners=[2419 26]; % right top of the canvas
                TobiiPoints_corners(2,:)=[1315 970]; % point at the bottom ring of the drum
                OriginalPoints_corners(2,:)=[2340 1570]; % point at the bottom ring of the drum
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 12
                TobiiPoints_corners=[540 85]; % left top of the canvas
                OriginalPoints_corners=[31 33]; % left top of the canvas
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 13
                TobiiPoints_corners=[355 835]; % left bottom of the painting
                OriginalPoints_corners=[57 1995]; % left bottom of the painting
                TobiiPoints_corners(2,:)=[390 8]; % nose of figure at the left (Reijnier Engelen)
                OriginalPoints_corners(2,:)=[152 902]; % nose of figure at the left (Reijnier Engelen)
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 14
                TobiiPoints_corners=[1671 916]; % right bottom of the canvas
                OriginalPoints_corners=[2413 2025]; % right bottom of the canvas
                TobiiPoints_corners(2,:)=[1494 46]; % tip of pike at the right (held by Walich Schellingwou)
                OriginalPoints_corners(2,:)=[2268 643]; % tip of pike at the right (held by Walich Schellingwou)
                TobiiPoints_corners(3,:)=[77 361]; % nose of second figure from the left (Jan Pietersen Bronckhorst)
                OriginalPoints_corners(3,:)=[329 1026]; % nose of second figure from the left (Jan Pietersen Bronckhorst)
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 15
                TobiiPoints_corners=[1208 128]; % right top of the canvas
                OriginalPoints_corners=[2419 26]; % right top of the canvas
                TobiiPoints_corners(2,:)=[1250 891]; % hole on drum
                OriginalPoints_corners(2,:)=[2283 1462]; % hole on drum
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 16
                TobiiPoints_corners=[501 235]; % left top of the canvas
                OriginalPoints_corners=[31 33]; % left top of the canvas
                TobiiPoints_corners(2,:)=[485 673]; % nose of figure at the left (Reijnier Engelen)
                OriginalPoints_corners(2,:)=[151 900]; % nose of figure at the left (Reijnier Engelen)
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 17
                TobiiPoints_corners=[596 815]; % left bottom of the painting
                OriginalPoints_corners=[57 1995]; % left bottom of the painting
                TobiiPoints_corners(2,:)=[626 51]; % nose of figure at the left (Reijnier Engelen)
                OriginalPoints_corners(2,:)=[151 900]; % nose of figure at the left (Reijnier Engelen)
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 18
                TobiiPoints_corners=[1222 729]; % right bottom of the canvas
                OriginalPoints_corners=[2413 2025]; % right bottom of the canvas
                TobiiPoints_corners(2,:)=[70 244]; % tip of leaf on the helmet of the shooting young musketeer
                OriginalPoints_corners(2,:)=[889 1205]; % tip of leaf on the helmet of the shooting young musketeer
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
        end
        TobiiPoints=[TobiiPoints;TobiiPoints_corners]; %#ok<AGROW>
        OriginalPoints=[OriginalPoints;OriginalPoints_corners]; %#ok<AGROW>
        
        if i==9 % output figure for paper (reference image 9)
            %  Figure 5. Feature matching results for an eye-tracker reference image (left) and the frontal image of The Night Watch (right). Three features were manually added (tip of musket, right eye of drummer, right corner of painting).
            figure;
            showMatchedFeatures(squeeze(FIXED(i,:,:)),original_grayscale,TobiiPoints(:,1:2),OriginalPoints,'montage');
            set(gca,'TickDir','out','LooseInset',[0.01 0.01 0.01 0.01])
            set(gca,'Position',[0.01 0.01 0.98 0.98])
        end
        
        Offsetvector=-6000:50:6000; % vector of possible values for b6 and b7
        CC=NaN(length(Offsetvector),length(Offsetvector),2);
        OriginalPoints_pred=NaN(length(Offsetvector),length(Offsetvector),size(TobiiPoints,1),2);
        bx=NaN(length(Offsetvector),length(Offsetvector),5);
        by=NaN(length(Offsetvector),length(Offsetvector),5);
        
        for xxi=1:length(Offsetvector) % loop over b6
            for yyi=1:length(Offsetvector) % loop over b7
                o1=TobiiPoints(:,2)-Offsetvector(yyi); % y-b7
                o2=TobiiPoints(:,1)-Offsetvector(xxi); % x-b6
                TobiiPoints=[TobiiPoints(:,1:2) o2./bound(o1,-150,150) o1./bound(o2,-150,150)]; % x, y, (x-b6)/(y-b7), (y-b7)/(x-b6)
                warning off all
                bx(xxi,yyi,:)=robustfit(TobiiPoints,OriginalPoints(:,1)); % perform regression analysis to predict x on the frontal image (Eq. 1)
                by(xxi,yyi,:)=robustfit(TobiiPoints,OriginalPoints(:,2)); % perform regression analysis to predict y on the frontal image (Eq. 2)
                warning on all
                disp([xxi yyi])
                OriginalPoints_pred(xxi,yyi,:,1)=bx(xxi,yyi,1)+bx(xxi,yyi,2)*TobiiPoints(:,1)+bx(xxi,yyi,3)*TobiiPoints(:,2)+bx(xxi,yyi,4)*TobiiPoints(:,3)+bx(xxi,yyi,5)*TobiiPoints(:,4); % predicted x on the frontal image (Eq. 1)
                OriginalPoints_pred(xxi,yyi,:,2)=by(xxi,yyi,1)+by(xxi,yyi,2)*TobiiPoints(:,1)+by(xxi,yyi,3)*TobiiPoints(:,2)+by(xxi,yyi,4)*TobiiPoints(:,3)+by(xxi,yyi,5)*TobiiPoints(:,4); % predicted y on the frontal image (Eq. 2)
                CC(xxi,yyi,:)=diag(corr(squeeze(OriginalPoints_pred(xxi,yyi,:,:)),OriginalPoints)); % correlation coefficient between points on the reference image and the frontal image
            end
        end
        [ovx_x,ovy_x]=find(CC(:,:,1)==max(max(CC(:,:,1))),1,'first'); % obtain indexes of maximum correlation coefficient between x-coordinates on the reference image and the frontal image
        [ovx_y,ovy_y]=find(CC(:,:,2)==max(max(CC(:,:,2))),1,'first'); % obtain indexes of maximum correlation coefficient between y-coordinates on the reference image and the frontal image
        
        if i==9 % output figure for paper (reference image 9)
            % Figure 6. Scatter plot of x_actual vs. x_predicted (left) and y_actual vs. y_predicted (right). The diagonal line is the line of unity. Markers (n = 335) are shown with transparency so that overlap can be seen more clearly. r > .9999 for the left and right panels.
            figure; subplot(1,2,1)
            scatter1=scatter(squeeze(OriginalPoints_pred(ovx_x,ovy_x,:,1)),OriginalPoints(:,1),100,'markerfacecolor','b','markeredgecolor','none');hold on;grid on
            plot([0 2500],[0 2500],'m:','Linewidth',2) % plot x_actual vs. x_predicted
            scatter1.MarkerFaceAlpha = .1;
            xlabel('Predicted \it{x}\rm-coordinate of feature')
            ylabel('Actual \it{x}\rm-coordinate of feature')
            set(gca,'pos',[0.06 0.07 0.40  0.9],'xlim',[0 2444],'ylim',[0 2444],'xtick',0:500:10000,'ytick',0:500:10000)
            xtickangle(0)
            subplot(1,2,2)
            scatter2=scatter(squeeze(OriginalPoints_pred(ovx_y,ovy_y,:,2)),OriginalPoints(:,2),100,'markerfacecolor','b','markeredgecolor','none');hold on;grid on
            scatter2.MarkerFaceAlpha = .1;
            plot([0 2500],[0 2500],'m:','Linewidth',2) % plot y_actual vs. y_predicted
            xlabel('Predicted \it{y}\rm-coordinate of feature')
            ylabel('Actual \it{y}\rm-coordinate of feature')
            set(gca,'pos',[0.56 0.07 0.40  0.9],'xlim',[0 2048],'ylim',[0 2048],'xtick',0:500:10000,'ytick',0:500:10000)
            xtickangle(0)
            h=findobj('FontName','Helvetica');set(h,'FontSize',20,'Fontname','Arial')
        end
        
        disp('Correlation between the actual and predicted horizontal coordinate, x')
        disp(corr(squeeze(OriginalPoints_pred(ovx_x,ovy_x,:,1)),OriginalPoints(:,1)))
        disp('Correlation between the actual and predicted vertical coordinate, y')
        disp(corr(squeeze(OriginalPoints_pred(ovx_y,ovy_y,:,2)),OriginalPoints(:,2)))
        
        RegressionCoefficients(i,1,:)=[squeeze(bx(ovx_x,ovy_x,:))' Offsetvector(ovx_x) Offsetvector(ovy_x)]; % store the 7 regression coefficients (bx1-7) that correspond to the maximum correlation
        RegressionCoefficients(i,2,:)=[squeeze(by(ovx_y,ovy_y,:))' Offsetvector(ovx_y) Offsetvector(ovy_y)]; % store the 7 regression coefficients (by1-7) that correspond to the maximum correlation
    end % end of reference images loop
    save('RegressionCoefficients.mat','RegressionCoefficients'); % 9 reference images x 2 coordinates (x,y) x 7 regression coefficients
else
    load RegressionCoefficients.mat
end
%% Figure S2. Top: Reconstructed frontal view from five reference images (see Figure 2) of the real Night Watch. The reconstruction was performed using Eqs. (1) and (2).
if recreate2dimage==1
    clear OriginalPoints_pred_All OriginalPoints_pred
    [Recreate2D,Recreate2Dc]=deal(NaN(size(original_grayscale,1),size(original_grayscale,2)));
    for i=1:5 % loop over five reference images
        bx=squeeze(RegressionCoefficients(i,1,:))'; % seven regression coefficients for image i and x-coordinate (Eq. 1)
        by=squeeze(RegressionCoefficients(i,2,:))'; % seven regression coefficients for image i and y-coordinate (Eq. 2)
        for II=1:1080 % loop over y (vertical) coordinates of reference image
            for JJ=1:1920 % loop over x (horizontal) coordinates of reference image
                
                o1=II-bx(7); % y-b7
                o2=JJ-bx(6); % x-b6
                TobiiPoints=[JJ II o2/bound(o1,-150,150) o1/bound(o2,-150,150)]; % x, y, (x-b6)/(y-b7), (y-b7)/(x-b6)
                OriginalPoints_pred(1)=bx(1)+bx(2)*TobiiPoints(1)+bx(3)*TobiiPoints(2)+bx(4)*TobiiPoints(3)+bx(5)*TobiiPoints(4); % predicted x on the frontal image (Eq. 1)
                
                o1=II-by(7); % y-b7
                o2=JJ-by(6); % x-b6
                TobiiPoints=[JJ II o2/bound(o1,-150,150) o1/bound(o2,-150,150)]; % x, y, (x-b6)/(y-b7), (y-b7)/(x-b6)
                OriginalPoints_pred(2)=by(1)+by(2)*TobiiPoints(1)+by(3)*TobiiPoints(2)+by(4)*TobiiPoints(3)+by(5)*TobiiPoints(4); % predicted y on the frontal image (Eq. 2)
                
                if i==1 % reference image 1
                    if OriginalPoints_pred(2) > 500 && OriginalPoints_pred(2) <= 1750 && OriginalPoints_pred(1) > 250 && OriginalPoints_pred(1) <= 2200 % central piece
                        if isnan(Recreate2D(round(OriginalPoints_pred(2)),round(OriginalPoints_pred(1)))) % only if pixel is not predicted yet
                            Recreate2D(round(OriginalPoints_pred(2)),round(OriginalPoints_pred(1)))=FIXED(i,II,JJ);
                        end
                    end
                end
                
                if i==2 % reference image 2
                    if  (OriginalPoints_pred(2) >=1 && OriginalPoints_pred(2) <= 500 && OriginalPoints_pred(1) >= 1222 && OriginalPoints_pred(1) <= 2444) || ... % right top
                            (OriginalPoints_pred(2) >500 && OriginalPoints_pred(2) < 1024 && OriginalPoints_pred(1) > 2200 && OriginalPoints_pred(1) <= 2444)    % right
                        if isnan(Recreate2D(round(OriginalPoints_pred(2)),round(OriginalPoints_pred(1)))) % only if pixel is not predicted yet
                            Recreate2D(round(OriginalPoints_pred(2)),round(OriginalPoints_pred(1)))=FIXED(i,II,JJ);
                        end
                    end
                end
                
                if i==3 % reference image 3
                    if  (OriginalPoints_pred(2) >=1 && OriginalPoints_pred(2) <= 500 && OriginalPoints_pred(1) >= 1 && OriginalPoints_pred(1) < 1222) || ... % left top
                            (OriginalPoints_pred(2) >500 && OriginalPoints_pred(2) < 1024 && OriginalPoints_pred(1) >= 1 && OriginalPoints_pred(1) <= 250)  % left
                        if isnan(Recreate2D(round(OriginalPoints_pred(2)),round(OriginalPoints_pred(1)))) % only if pixel is not predicted yet
                            Recreate2D(round(OriginalPoints_pred(2)),round(OriginalPoints_pred(1)))=FIXED(i,II,JJ);
                        end
                    end
                end
                
                if i==4 % reference image 4
                    if  (OriginalPoints_pred(2) >=1750 && OriginalPoints_pred(2) <= 2048 && OriginalPoints_pred(1) >= 1 && OriginalPoints_pred(1) < 1222) || ... % left bottom
                            (OriginalPoints_pred(2) >= 1024 && OriginalPoints_pred(2) < 2048 && OriginalPoints_pred(1) >= 1 && OriginalPoints_pred(1) <= 250)  % left
                        if isnan(Recreate2D(round(OriginalPoints_pred(2)),round(OriginalPoints_pred(1)))) % only if pixel is not predicted yet
                            Recreate2D(round(OriginalPoints_pred(2)),round(OriginalPoints_pred(1)))=FIXED(i,II,JJ);
                        end
                    end
                end
                
                if i==5 % reference image 5
                    if  (OriginalPoints_pred(2) >=1750 && OriginalPoints_pred(2) <= 2048 && OriginalPoints_pred(1) >= 1222 && OriginalPoints_pred(1) < 2444) || ... % right bottom
                            (OriginalPoints_pred(2) >= 1024 && OriginalPoints_pred(2) < 2048 && OriginalPoints_pred(1) >= 2200 && OriginalPoints_pred(1) <= 2444)  % right
                        if isnan(Recreate2D(round(OriginalPoints_pred(2)),round(OriginalPoints_pred(1)))) % only if pixel is not predicted yet
                            Recreate2D(round(OriginalPoints_pred(2)),round(OriginalPoints_pred(1)))=FIXED(i,II,JJ);
                        end
                    end
                end
                
            end % end of x (horizontal) coordinates loop
        end % end of y (vertical) coordinates loop
    end % end of reference images loop
    
    [ii,jj]=find(isnan(Recreate2D)); % find pixels that were not predicted
    
    for i=1:length(ii) % loop over pixels that were not predicted, to fill with neighboring values
        if ii(i)>1&&jj(i)>1&&ii(i)<size(Recreate2D,1)&&jj(i)<size(Recreate2D,2)
            pixs=Recreate2D(ii(i)-1:ii(i)+1,jj(i)-1:jj(i)+1);
            Recreate2D(ii(i),jj(i))=mean(pixs(:),'omitnan');
        end
    end
    Recreate2D(isnan(Recreate2D))=0;
    save('Recreate2D.mat','Recreate2D') % save frontal image of the painting recreated from reference images from the real Night Watch experiment
end
%% This block matches the Tobii image segment being gazed at with the images
%% A loop is run over all participants and video frames
%% The results are outputted as x and y coordinates of the matched reference image (participant x frame x reference image x window size)
%% The results are saved in CData6... .mat
if match_to_reference_images==1
    [MaxCor,II,JJ]=deal(NaN(21,15000,9,length(window_size)));
    Coordinates_store=NaN(21,15000,2);
    FIXED_GPU=gpuArray(FIXED);
    
    close all
    for p=1:21 % loop over participants
        Xp=squeeze(X(p,:,:)); % select data for participant p
        Xp=Xp(Xp(:,52)==1,:); % only select data points for which video recording is available
        
        videoname=mapping{p,3};
        if ismember(p,17:21) % measurement day 2
            vidObj=VideoReader(['..\..\Experiment - Rijksmuseum - Take2\Project1\Media\' videoname]);
        else
            vidObj=VideoReader(['..\Project1\Media\' videoname]);
        end
        
        Time=Xp(:,1);
        Time=Time/10^6-Time(end)/10^6+vidObj.Duration; % create time vector that matches the video time
        
        eyetracker=find(Xp(:,51)==2 & Time<=max(Time)-0.5); % find indexes of eye tracker data and discard last 0.5 seconds
        
        Xp=Xp(eyetracker,:); % select only the eye tracker data
        Gaze=Xp(:,16:17); % gaze point x, gaze point y (in pixels)
        Time=Time(eyetracker,:); % time vector
        FrameNumber=ceil(Time*25.01762); % compute video frame number that corresponds to gaze data
        
        % compute miss_start and miss_end; they define the gaps in the data, for example due to blinks
        miss=isnan(Gaze(:,1)); % missing eye tracking data
        miss_start=1+find(diff(miss)>0); % start indexes of intervals with NaNs
        miss_end=NaN(size(miss_start));
        for i=1:length(miss_start)
            try
                miss_end(i)=miss_start(i)+find(Gaze(miss_start(i):end,1)>0,1,'first')-2; % end indexes of intervals with NaNs
            catch error
            end
        end
        
        GazeC=Gaze;
        for i=1:length(miss_start) % increase the gap sizes with 2 frames. This is done because the eyelid opening and closing take about 2 frames each. These data need to be interpolated
            try
                GazeC(miss_start(i)-2:miss_end(i)+2)=NaN;
            catch error
            end
        end
        GazeCF=NaN(size(GazeC));
        GazeCF(:,1)=fixgaps(GazeC(:,1)); % linearly interpolate eye gaze gaps
        GazeCF(:,2)=fixgaps(GazeC(:,2)); % linearly interpolate eye gaze gaps
        GazeCF=movmedian(GazeCF,20); % median filter to cancel out noise
        
        for i=1:length(miss_start)
            if miss_end(i)-miss_start(i)>50 % if the gap is greater than 0.5 s
                GazeCF(miss_start(i):miss_end(i),:)=NaN; % then set the data to NaN. We do not want long gaps (> 0.5 s) to be interpolated as this would be misleading (suggesting there is data while there is none)
            end
        end
        
        vr = VideoWriter(['Moving_with_marker_participant_no_frameno_' num2str(p)], 'Mpeg-4'); %#ok<*TNMLP>
        vr.Quality=100;vr.FrameRate=25;
        open(vr);
        begin=1; % can be changed if wanting to show a specific frame
        
        for f=begin:max(FrameNumber) % loop over frames to be analysed
            disp(['Coordinates of the match of the Tobii image segment being gazed at with the reference images, Participant number = ' num2str(p) ', Frame number = ' num2str(f)])
            
            if rem(f,100)==1 || f==begin % loading batches of 100 frames, to speed up the code
                fend=f+99;
                if fend>vidObj.NumFrames
                    fend=vidObj.NumFrames;
                end
                moving = read(vidObj, [f fend]); % read batches of the video, e.g. 1-100, 101-200, etc
                fbegin=f;
            end
            moving_with_marker=moving(:,:,:,f-fbegin+1);
            moving_gray=rgb2gray(squeeze(moving(:,:,:,f-fbegin+1))); % 1080 x 1920 matrix
            eye_frame=find(FrameNumber==f); % eye-gaze indexes of frame
            if ~isempty(eye_frame) && ~isnan(nanmean(GazeCF(eye_frame,1)))  %#ok<NANMEAN> % if there is data for this frame f
                GazeFrame=nanmean(GazeCF(eye_frame,:),1); %#ok<NANMEAN> % gaze x and y corresponding to frame f
                x=round(GazeFrame(:,1)); % x gaze coordinate in Tobii image; round to nearest integer (pixel)
                y=round(GazeFrame(:,2)); % y gaze coordinate in Tobii image; round to nearest integer (pixel)
                moving_with_marker = insertShape(moving_with_marker,'FilledCircle',[x y 15],'color',[0 255 0],'opacity',1); % add circular green marker to the Tobii video
                Coordinates_store(p,f,:)=[x y]; % x, y gaze coordinates
                
                for ws=1:length(window_size)
                    
                    AOI=gpuArray(imcrop(moving_gray,[x-window_size(ws) y-window_size(ws) 2*window_size(ws) 2*window_size(ws)])); % rectangle around gaze point (e.g., if window_size(ws)=300, then AOI is 601 x 601 pixels)
                    for i=1:9 % loop over reference images
                        if std(std(double(AOI)))>0 % there needs to be at least some variation in the AOI (not only white pixels by looking at the reflection)
                            if ismember(p,17:21)
                                C=normxcorr2(AOI, squeeze(FIXED_GPU(i+9,:,:))); % normalized two-dimensional cross-correlation, use the second set of reference images for participants 17-21. These participants were measured on a different day
                            else
                                C=normxcorr2(AOI, squeeze(FIXED_GPU(i,:,:))); % normalized two-dimensional cross-correlation
                            end
                            MaxCor(p,f,i,ws)=max(C(:)); % find maximum correlation between AOI and reference image
                            [II(p,f,i,ws),JJ(p,f,i,ws)]=find(C==MaxCor(p,f,i,ws),1,'first'); %  II: y coordinate, JJ: x coordinate
                            II(p,f,i,ws)=II(p,f,i,ws)-window_size(ws); % compute the y gaze coordinate in the reference image of maximal correlation
                            JJ(p,f,i,ws)=JJ(p,f,i,ws)-window_size(ws); % compute the x gaze coordinate in the reference image of maximal correlation
                            
                            if y>1080-window_size(ws) % if the participant gazed at the bottom of the Tobii image, then apply correction to account for cropping effects
                                II(p,f,i,ws)=II(p,f,i,ws)+y-1080+window_size(ws);
                            end
                            if x>1920-window_size(ws) % if the participant gazed at the right of the Tobii image, then apply correction to account for cropping effects
                                JJ(p,f,i,ws)=JJ(p,f,i,ws)+x-1920+window_size(ws);
                            end
                            
                            % Example verification
                            % participant 1, frame 202, window_size(ws) = 300
                            % Tobii gaze coordinates 272, 1630 (y, x)
                            % so the participant gazed at the upper right, which is why the AOI is cropped at the top and at the right
                            % size(AOI): 572 (1:572), 591 (1330:1920) (y, x)
                            
                            % the size of C (output of normmxcor2) = size(moving_gray)+size(AOI)-1
                            % coordinate of maximum correlation according to normxcorr2, including padding: i = 1080, j = 2354 (y, x)
                            % so the left top of the match is at 508, 1763 in the matrix excluding padding (y, x)
                            % Thus, the coordinates of the gaze in the reference image are
                            % 779, 2063 (y, x)
                            
                            % AOI size subtracted (per https://nl.mathworks.com/help/images/ref/normxcorr2.html)
                            % yoffSet = ypeak-size(onion,1);xoffSet = xpeak-size(onion,2); to account for padding of normxcorr2
                            % i-size(AOI,1) = 508; j-size(AOI,2) = 1763 (y, x; so this is the left top of the match)
                            
                        end
                    end % end of reference images loop
                end % end of window sizes loop
            end
            %% (frame = 3000, participant = 1)
            %% Figure 4. Left top: Eye-tracker camera image (1920x1080 pixels) for a random frame with a marker drawn at the measured gaze coordinate. Middle top: Heatmap of correlation coefficients (1920x1080 pixels) between one of the templates in grayscale and one of the reference images in grayscale (1920x1080 pixels). Right top: Estimated gaze positions, i.e., locations of the maximal correlation coefficient, in one of the reference images (here zoomed in to 242x136 pixels, for clarity). Seven markers are shown, corresponding to the seven templates. The markers are color-coded according to the maximal correlation coefficient (r = .72, .73, .83, .82, .87, .90, .93, respectively). Bottom: Templates in grayscale of seven different sizes centered around the gaze coordinate in the eye-tracker camera image. The templates are square-shaped unless they are out of bounds.
            if plot_for_participant==1 && f==3000 && p==1
                figure;
                subplot(2,7,1)
                imshow(moving_with_marker);
                set(gca,'pos',[0.00 0.46 0.3 0.3])
                title('Eye-tracker image with gaze marker')
                [MaxCorp,IIp,JJp]=deal(NaN(length(window_size),1));
                
                for ws2=1:length(window_size) % loop over window sizes
                    x=round(GazeFrame(:,1)); % x gaze coordinate in Tobii image; round to nearest integer (pixel)
                    y=round(GazeFrame(:,2)); % y gaze coordinate in Tobii image; round to nearest integer (pixel)
                    AOI=imcrop(moving_gray,[x-window_size(ws2) y-window_size(ws2) 2*window_size(ws2) 2*window_size(ws2)]); % rectangle around gaze point (e.g., if window_size(ws)=300, then AOI is 601 x 601 pixels)
                    C=normxcorr2(AOI, squeeze(FIXED_GPU(5,:,:))); % use the second set of reference images for p17-21. These participants were measured on a different day
                    
                    MaxCorp(ws2)=max(C(:));
                    [IIp(ws2),JJp(ws2)]=find(C==MaxCorp(ws2),1,'first'); % II = y coordinate, JJ = x coordinate
                    IIp(ws2)=IIp(ws2)-window_size(ws2); % compute the y gaze coordinate in the reference image
                    JJp(ws2)=JJp(ws2)-window_size(ws2); % compute the x gaze coordinate in the reference image
                    
                    if y>1080-window_size(ws2) % if the participant gazed at the bottom of the Tobii image
                        IIp(ws2)=IIp(ws2)+y-1080+window_size(ws2);
                    end
                    if x>1920-window_size(ws2) % if the participant gazed at the right of the Tobii image
                        JJp(ws2)=JJp(ws2)+x-1920+window_size(ws2);
                    end
                    
                    subplot(2,7,ws2+7)
                    imshow(AOI) % show cropped images per window size at the bottom
                    title([num2str(size(AOI,2)) 'x'  num2str(size(AOI,1)) ' pixels'])
                    set(gca,'pos',[(ws2-1)/7.6 0.22 0.155 0.155])
                    
                    if ws2==7 % plot correlation heatmap for smallest AOI template
                        subplot(2,7,3)
                        imshow(C);
                        
                        colormap(gca,mycolormap)
                        title('2D correlations (for 101x101-pixel template)')
                        colorbar;
                        set(gca,'pos',[0.305 0.46 0.3 0.3])
                        
                        subplot(2,7,7)
                        reference_image_with_marker=squeeze(FIXED_RGB(5,:,:,:));
                        JJp=gather(JJp);IIp=gather(IIp);
                        for ws3=1:7
                            reference_image_with_marker = insertShape(reference_image_with_marker,'Circle',[JJp(ws3) IIp(ws3) 8],'color',255*mycolormap(ceil(MaxCorp(ws3)*256),:),'opacity',1); % add circular green marker
                            mycolormap(ceil(MaxCorp(ws3)*256),:)
                        end
                        imshow(reference_image_with_marker)
                        colormap(gca,mycolormap)
                        colorbar
                        title('Estimated position reference image')
                        set(gca,'pos',[0.66 0.46 0.3 0.3]);
                        set(gca,'xlim',[714 1011],'ylim',[163 331])
                    end
                end
                h=findobj('FontName','Helvetica'); set(h,'FontSize',17,'Fontname','Arial');
                set(gca,'TickDir','out','LooseInset',[0.01 0.01 0.01 0.01])
                disp('Correlation coefficients Figure 4, right top')
                disp(round(gather(MaxCorp),2))
            end
        end % end of frames loop
        close(vr)
        filename=['Cdata6_' num2str(round(now*36*2400)) '.mat'];
        save(filename,'MaxCor','II','JJ','Coordinates_store')
    end % end of participants loop
else
    load Cdata6_63799705098.mat
end
%% This block computes the points on the frontal image of the painting with the help of the regression coefficients
%% The XY coordinates are stored in a matrix 'OriginalPoints_pred_All' (participants x frames x reference image no x window size)
clear OriginalPoints_pred_All OriginalPoints_pred
OriginalPoints_pred_All=NaN(size(II,1),size(II,2),size(II,3),size(II,4),2);
for p=1:21 % loop over participants
    disp(['Compute points on original painting based on regression coefficients, Participant number = ' num2str(p)])
    for i=1:9 % loop over reference images
        if ismember(p,17:21) % measurement day 2
            bx=squeeze(RegressionCoefficients(i+9,1,:))';
            by=squeeze(RegressionCoefficients(i+9,2,:))';
        else % measurement day 1
            bx=squeeze(RegressionCoefficients(i,1,:))';
            by=squeeze(RegressionCoefficients(i,2,:))';
        end
        for i2=1:7 % loop over AOI window sizes
            for s=1:size(II,2) % loop over frames
                if II(p,s,i,i2) >= 1 && JJ(p,s,i,i2) >= 1 && II(p,s,i,i2) <= 1080 && JJ(p,s,i,i2) <= 1920 % only make the prediction within bounds
                    o1=II(p,s,i,i2)-bx(7); % y-b7
                    o2=JJ(p,s,i,i2)-bx(6); % x-b6
                    TobiiPoints=[JJ(p,s,i,i2) II(p,s,i,i2) o2/bound(o1,-150,150) o1/bound(o2,-150,150)]; % x, y, (x-b6)/(y-b7), (y-b7)/(x-b6)
                    OriginalPoints_pred(1)=bx(1)+bx(2)*TobiiPoints(1)+bx(3)*TobiiPoints(2)+bx(4)*TobiiPoints(3)+bx(5)*TobiiPoints(4); % predicted x on the frontal image (Eq. 1)
                    
                    o1=II(p,s,i,i2)-by(7); % y-b7
                    o2=JJ(p,s,i,i2)-by(6); % y-b7
                    TobiiPoints=[JJ(p,s,i,i2) II(p,s,i,i2) o2/bound(o1,-150,150) o1/bound(o2,-150,150)]; % x, y, (x-b6)/(y-b7), (y-b7)/(x-b6)
                    OriginalPoints_pred(2)=by(1)+by(2)*TobiiPoints(1)+by(3)*TobiiPoints(2)+by(4)*TobiiPoints(3)+by(5)*TobiiPoints(4); % predicted y on the frontal image (Eq. 2)
                    
                    OriginalPoints_pred_All(p,s,i,i2,1:2)=OriginalPoints_pred; % predict x and y
                end
            end
        end
    end
end
%% Compute stability scores for each reference image
%% The stability score is computed over a sliding window for the different window sizes
%% A high stability score is obtained when distances are relatively
%% constant, i.e., when the same location on the frontal image is identified
%% The stability score is helpful for determining the best match among window sizes and reference images. It compares coordinate estimates of subsequent window sizes for a given reference image
[stability_score,stability_score_index]=deal(NaN(size(II,1),size(II,2),size(II,3)));
wl=2; % sliding window length
for p=1:21 % loop over participants
    disp(['Compute stability scores for each reference image, Participant number = ' num2str(p)])
    for f=1:size(OriginalPoints_pred_All,2) % loop over frames
        for i=1:9 % loop over reference images
            OriginalPoints_pred_Sample_x=squeeze(OriginalPoints_pred_All(p,f,i,:,1));
            OriginalPoints_pred_Sample_y=squeeze(OriginalPoints_pred_All(p,f,i,:,2));
            
            ssd=NaN(size(II,4)-1,1);
            for i2=1:size(II,4)-1 % loop over window sizes
                ssd(i2)=pdist([OriginalPoints_pred_Sample_x(i2:i2+1) OriginalPoints_pred_Sample_y(i2:i2+1)]); % subsequent distance between predicted coordinates on the frontal image
            end
            ssd2=NaN(size(II,4)-wl,1);
            for i2=1:length(ssd)-wl+1
                ssd2(i2)=mean(ssd(i2:i2+wl-1)); % mean stability over wl points (lower means more consistent prediction)
            end
            [stability_score(p,f,i),stability_score_index(p,f,i)]=min(ssd2); % start index of stability interval for reference image i
        end
    end
end
%% Determining the final coordinate on the frontal image of the painting
%% From all candidate matching points on the frontal image identified above, the most appropriate one is selected based on stability and correlation coefficients
%% The result is XX and YY, i.e., the actual gaze point on the frontal image of the painting
[XX,YY]=deal(NaN(size(II,1),size(II,2)));
SelectedPoints=deal(NaN(size(II,1),size(II,2),2));
for p=1:21 % loop over participants
    disp(['Select the most appropriate point on the frontal image of the painting, Participant number = ' num2str(p)])
    for f=1:size(II,2) % loop over frames
        x=squeeze(OriginalPoints_pred_All(p,f,:,:,1));
        y=squeeze(OriginalPoints_pred_All(p,f,:,:,2));
        
        MaxCori=NaN(size(MaxCor,3),1);
        for i=1:size(MaxCor,3) % loop over reference images
            MaxCori(i)=squeeze(MaxCor(p,f,i,squeeze(stability_score_index(p,f,i)))); % look up the correlation coefficient of the template match that corresponds to the template size that yielded the best stability
        end
        ranks=-MaxCori+squeeze(stability_score(p,f,:))/100; % combine correlation coefficient and stability score using a heuristic weighting (lower score is better)
        [~,b]=min(ranks); % reference image number of best match between the template and the reference image
        b2=stability_score_index(p,f,b); % corresponding template size index
        XX(p,f)=median(x(b,b2:b2+wl)); % x-coordinate of gaze point on the frontal image of the painting
        YY(p,f)=median(y(b,b2:b2+wl)); % y-coordinate of gaze point on the frontal image of the painting
        SelectedPoints(p,f,:)=[b b2]; % reference image number and template size number of best match
    end
end

if plot_for_participant==1
    %% This block shows a selected frame of a selected participant
    %% This block does not have a fundamental role in creating the results but can be used for checking the algorithm
    clc;close all
    
    p=1;f=3000;
    
    videoname=mapping{p,3};
    x=squeeze(OriginalPoints_pred_All(p,f,:,:,1));
    disp('Example of a selected frame of a selected participant')
    disp('Coordinate x')
    disp(round(x))
    
    y=squeeze(OriginalPoints_pred_All(p,f,:,:,2));
    disp('Coordinate y')
    disp(round(y))
    disp('Maxcor')
    disp(squeeze(MaxCor(p,f,:,:)))
    disp('Maxcor, stability score, ranks, stability score index')
    disp([median(squeeze(MaxCor(p,f,:,stability_score_index(p,f,:))),2,'omitnan') squeeze(stability_score(p,f,:)) ranks squeeze(stability_score_index(p,f,:)) ])
    disp('Selected coordinates')
    disp([XX(p,f) YY(p,f)])
    disp('Selected points')
    disp(squeeze(SelectedPoints(p,f,:)))
    
    vidObj=VideoReader(['..\..\Experiment - Rijksmuseum\Project1\Media\' videoname]);
    
    moving = read(vidObj, f); % read batches of the video, e.g. 1-100, 101-200, etc
    moving_with_marker = insertShape(moving,'FilledCircle',[Coordinates_store(p,f,1) Coordinates_store(p,f,2) 15],'color',[0 255 0],'opacity',1); % add circular green marker to the Tobii video
    
    figure; % show Tobii frame with marker
    imshow(moving_with_marker)
    axis equal
    
    set(gca,'LooseInset',[0.01 0.01 0.01 0.01])
    
    original_with_marker=original;
    for i=1:size(x,1)
        for j=1:size(x,2)
            if ~isnan(x(i,j))&&~isnan(y(i,j))
                original_with_marker = insertShape(original_with_marker,'FilledCircle',[x(i,j) y(i,j) 15],'color',255*[1 1 1],'opacity',1); % add circular green marker to the Tobii video
                original_with_marker = insertText(original_with_marker,[x(i,j) y(i,j)],[num2str(round(x(i,j))) ', ' num2str(round(y(i,j)))]); % add circular green marker to the Tobii video
            end
        end
    end
    original_with_marker = insertShape(original_with_marker,'FilledCircle',[XX(p,f) YY(p,f) 10],'color','red','opacity',1); % add circular green marker to the Tobii video
    
    figure; % show original picture with markers
    imshow(original_with_marker)
end
%% Finally, after processing all frames, a median filter with a window length of 5 (0.2 s) is applied to the final x and y coordinates to filter out potential outliers
% Also out-of-bounds matches are set to NaN
XXf=medfilt1(XX',5,'omitnan')';
YYf=medfilt1(YY',5,'omitnan')';
XXf(XXf<0-100)=NaN;
XXf(XXf>2444+100)=NaN;
YYf(YYf<0-100)=NaN;
YYf(YYf>2048+100)=NaN;
startindex=[1313 1228 1441 1415 1195 1456 1018 793 1170 1042 1286 983 1462 831 1208 1029 1568 1166 4031 2688 6449]; % frame numbers per participant when the eye-gaze first lands on the painting
%% Create videos
%% 1. Create video with gaze markers for a single participant on the frontal image of the painting
if video_separate==1
    original_black=zeros(size(original,2),size(original,1));
    for p=1
        vr = VideoWriter(['MappingR3_' num2str(p)], 'Mpeg-4'); %#ok<*TNMLP>
        vr.Quality=100;vr.FrameRate=25;
        open(vr);
        
        videoname=mapping{p,3};
        
        if ismember(p,17:21) % measurement day 2
            vidObj=VideoReader(['..\..\Experiment - Rijksmuseum - Take2\Project1\Media\' videoname]);
        else % measurement day 1
            vidObj=VideoReader(['..\Project1\Media\' videoname]);
        end
        
        for f=startindex(p):startindex(p)+7500
            disp(['Mapping illustration for a single participant, Frame number = ' num2str(f)])
            moving = read(vidObj, f); % read batches of the video, e.g. 1-100, 101-200, etc.
            if ~isnan(Coordinates_store(p,f,1))&&~isnan(Coordinates_store(p,f,2))
                moving_with_marker = insertShape(moving,'FilledCircle',[Coordinates_store(p,f,1) Coordinates_store(p,f,2) 25],'color',[0 255 0],'opacity',1); % add circular green marker to the Tobii video
            end
            clear R
            R(151:150+size(moving_with_marker,1),:,:)=moving_with_marker;
            R=insertText(R,[10 10],['Frame number = ' num2str(f-startindex(p)+1)],'fontsize',30,'BoxColor','black','BoxOpacity',0,'TextColor','green');
            R=insertText(R,[10 50],['Time = ' sprintf('%0.2f',(f-startindex(p))/25) ' s'],'fontsize',30,'BoxColor','black','BoxOpacity',0,'TextColor','green');
            R=insertText(R,[620 1],'Camera on head','fontsize',90,'BoxColor','black','BoxOpacity',0,'TextColor','green');
            original_with_marker=original;
            if ~isnan(XXf(p,f))&&~isnan(YYf(p,f))
                original_with_marker= insertShape(original,'FilledCircle',[XXf(p,f) YYf(p,f) 30],'color',[0 255 0],'opacity',1); % add circular green marker to the Tobii video
            end
            original_with_marker=imresize(original_with_marker,0.7);
            R(126:125+size(original_with_marker,1),2001:2000+size(original_with_marker,2),:)=original_with_marker;
            R=insertText(R,[2450 1],'Frontal image','fontsize',90,'BoxColor','black','BoxOpacity',0,'TextColor','green');
            writeVideo(vr,R);
        end
        
        close(vr);
    end
end
%% 2. Create video with gaze markers of all participants on the frontal image of the painting
if video_combined==1
    vr = VideoWriter('Meta-video', 'Mpeg-4'); %#ok<*TNMLP>
    vr.Quality=100;vr.FrameRate=25;
    open(vr);
    
    Rcircle=15;
    original_black=zeros(size(original,2),size(original,1));
    original_with_marker = insertShape(original_black,'FilledCircle',[50,50,Rcircle],'color',[0 255 0],'opacity',1); % create circular green marker
    Circle=double(original_with_marker(50-Rcircle:50+Rcircle,50-Rcircle:50+Rcircle,2));
    Circle=uint8(Circle/1.5);
    
    for f=1:7500 % 5 minutes of data
        original_with_marker=original;
        disp(['Video creation with markers of all persons, Frame number = ' num2str(f)])
        for p=1:21
            fillx=round(XXf(p,f+startindex(p)-1)-Rcircle:XXf(p,f+startindex(p)-1)+Rcircle);
            filly=round(YYf(p,f+startindex(p)-1)-Rcircle:YYf(p,f+startindex(p)-1)+Rcircle);
            if isempty(find(~isnan(fillx)==0,1))&&isempty(find(fillx<=0,1))&&isempty(find(fillx>2444,1))&&isempty(find(~isnan(filly)==0,1))&&isempty(find(filly<=0,1))&&isempty(find(filly>2048,1))
                original_with_marker(filly,fillx,2) = original_with_marker(filly,fillx,2)+Circle;
            end
        end
        writeVideo(vr,original_with_marker);
    end
    
    close(vr);
end
%% This block creates the heatmap. It does so by adding a patch for each participant and frame
% Figure 9. Heatmap of the 21 participants combined. The heatmap was normalized by dividing by the number of patches and multiplying by 1000.
% Figure 10. Left: Heatmap for the first 60 seconds of viewing. Right: Heatmap for the last 60 seconds of viewing. Note that the color bar ranges from 0 to 18 but from 0 to 10 in the previous heatmap.

close all
Rcircle=30;
original_with_marker = insertShape(zeros(500,500,3),'FilledCircle',[250,250,Rcircle],'color',[0 255 0],'opacity',1); % add circular green marker to black background
Circle=double(original_with_marker(250-Rcircle:250+Rcircle,250-Rcircle:250+Rcircle,2)); % cut out the circle
Circle=Circle/255; % we are left with a circle of radius Rcircle

HMstore=NaN(66,2048,2444); % 21 participants x 3 + 3
for hm=1:size(HMstore,1) % create 3 heatmaps and heatmaps for individual participants
    switch hm
        case 1
            IDs=1:21; % all participants
            V=1:7500; % entire 5 minutes
        case 2
            IDs=1:21; % all participants
            V=1:1500; % first minute
        case 3
            IDs=1:21; % all participants
            V=6001:7500; % last minute
    end
    if ismember(hm,4:24)
        IDs=hm-3; % create heatmaps for individual participants
        V=1:7500; % entire 5 minutes
    end
    if ismember(hm,25:45)
        IDs=hm-24; % create heatmaps for individual participants
        V=1:1500; % first minute
    end
    if ismember(hm,46:66)
        IDs=hm-45; % create heatmaps for individual participants
        V=6001:7500; % last minute
    end
    
    original_black=zeros(size(original,2),size(original,1));
    
    if hm==1 % if first heatmap, with all participants
        [XA,YA,XAf,YAf]=deal(NaN(21,7501));
    end
    counter=0;
    for i=1:length(IDs) % loop over all included participants
        p=IDs(i);
        disp(['Heatmap creation, Participant number = ' num2str(p)])
        if hm==1
            XA(p,:)=XX(p,startindex(p):startindex(p)+7500); % store 5 minutes of data since the starting moment of the trial
            YA(p,:)=YY(p,startindex(p):startindex(p)+7500);
            XAf(p,:)=XXf(p,startindex(p):startindex(p)+7500);
            YAf(p,:)=YYf(p,startindex(p):startindex(p)+7500);
        end
        for f=V % loop over frames
            fillx=XXf(p,f+startindex(p)-1)-Rcircle:XXf(p,f+startindex(p)-1)+Rcircle; % create x coordinate range for frame f
            filly=YYf(p,f+startindex(p)-1)-Rcircle:YYf(p,f+startindex(p)-1)+Rcircle; % create y coordinate range for frame f
            fillxr=round(fillx); % x gaze coordinate in Tobii image; round to nearest integer (pixel)
            fillyr=round(filly); % y gaze coordinate in Tobii image; round to nearest integer (pixel)
            if isempty(find(~isnan(fillxr)==0,1))&&isempty(find(fillxr<=0,1))&&isempty(find(fillxr>2444,1))&&isempty(find(~isnan(fillyr)==0,1))&&isempty(find(fillyr<=0,1))&&isempty(find(fillyr>2048,1))
                original_black(fillxr,fillyr)=original_black(fillxr,fillyr)+Circle; % add the circular patch for each frame (5 minutes of data since the starting moment) and each participant
                counter=counter+1;
            end
        end
    end
    HMstore(hm,:,:)=1000*original_black'./counter; % heatmap values
    
    % Here we display the heatmap
    if ismember(hm,1:3)
        opengl software
        original_grayscale3(:,:,1)=original_grayscale;
        original_grayscale3(:,:,2)=original_grayscale;
        original_grayscale3(:,:,3)=original_grayscale;
        figure;
        imshow(original_grayscale3);hold on % frontal image of the painting
        OverlayImage = imshow(1000*original_black'./counter);
        set(OverlayImage, 'AlphaData',0.6)
        
        colormap(mycolormap);
        disp(['Maximum heatmap value: ' num2str(max(max(1000*original_black./counter)))])
        colorbar
        axis equal
        if hm==1
            caxis([0 10]);
        else
            caxis([0 18]);
        end
        set(gca,'TickDir','out','LooseInset',[0.01 0.01 0.01 0.01])
        h=findobj('FontName','Helvetica'); set(h,'FontSize',24,'Fontname','Arial');
        set(gca,'TickDir','out','LooseInset',[0.01 0.01 0.01 0.01])
        set(gca,'Position',[0.00 0.03 0.94 0.94])
    end
end
%% Compute local entropy
EntropyImg=double(original_grayscale)/255; % normalize from 0 to 1
EntropyImg=entropyfilt(EntropyImg,true(15));

% Figure 1. Bottom: Saliency map consisting of the local entropy computed for the upper image in grayscale.
figure; imshow(EntropyImg)
h=findobj('FontName','Helvetica'); set(h,'FontSize',24,'Fontname','Arial');
set(gca,'TickDir','out','LooseInset',[0.01 0.01 0.01 0.01])
mycolormap=colormap("hsv");
ft=customcolormap([0 1],[mycolormap(52,:);0.5 0.5 0.5]);
mycolormap(1:52,:)=ft(1:5:end,:);
colorbar;
colormap(mycolormap);
caxis([0 7.8])
h=findobj('FontName','Helvetica'); set(h,'FontSize',24,'Fontname','Arial');
set(gca,'TickDir','out','LooseInset',[0.01 0.01 0.01 0.01])
set(gca,'Position',[0.00 0.03 0.94 0.94])
%%
original_nose=original;
noses=[151 900 % x and y coordinates of all visible characters' noises in the frontal image of the paining
    225 1380
    245 920
    327 1024
    364 1007
    439 888
    552 1085
    725 1538
    740 855
    802 1260
    848 813
    979 912
    983 1202
    1059 854
    1195 887
    1201 1041
    1266 870
    1323 1063
    1343 896
    1429 1136
    1501 840
    1640 1204
    1790 1044
    1971 1033
    2070 1690
    2231 948
    2296 1137
    2301 945];
Eccentricity=1.895/1000*pdist([[2444/2 2048/2 ];noses]);Eccentricity=Eccentricity(1:28); % eccentricity was defined as the distance in meters between the AOI’s landmark point and the center of the painting
colors=distinguishable_colors(size(noses,1));
rng(1);[~,o]=sort(randn(size(noses,1),1));colors=colors(o,:); % sort the colors in a random order

% Figure 7. Voronoi cells around all characters’ noses. These cells were used as areas of interest.
figure;hold on
mask=voronoi2mask(noses(:,1),noses(:,2),size(original_grayscale));hold on
maskc = label2rgb(mask, colors);

imshow(original_grayscale3);hold on % frontal image of the painting
OverlayImage = imshow(maskc);
set(OverlayImage, 'AlphaData',0.6)
set(gca,'TickDir','out','LooseInset',[0.01 0.01 0.01 0.01])
h=findobj('FontName','Helvetica'); set(h,'FontSize',24,'Fontname','Arial');
set(gca,'TickDir','out','LooseInset',[0.01 0.01 0.01 0.01])
set(gca,'Position',[0.00 0.03 0.94 0.94])
set(gcf,'color','w');

for i=1:size(noses,1)
    text(noses(i,1),noses(i,2),num2str(i),'color','white','fontsize',20,'horizontalalignment','center');
end
load('..\Experiment Replica\HMstore_replica.mat') % load heatmap data from replica Night Watch experiment (n = 27)
load('..\Experiment Replica\Recreate2D_replica.mat') % frontal image recreated from reference image, from replica Night Watch experiment
load('Recreate2D.mat') % frontal image recreated from reference image, real Night Watch experiment

% Figure S2. Top: Reconstructed frontal view from five reference images (see Figure 2) of the real Night Watch. Bottom: Reconstructed frontal view from a reference image of the replica Night Watch (bottom) (see Figure 8).
figure;imshow((Recreate2D/255))
figure;imshow((Recreate2D_replica/255))

Recreate2D_replica(1:100,:)= 0; % create black borders for replica image
Recreate2D_replica(size(Recreate2D_replica,1)-55+1:end,:)=0;
Recreate2D_replica(:,1:55)=0;
Recreate2D_replica(:,size(Recreate2D_replica,2)-65+1:end)=0;

VM=NaN(size(noses,1),10); % 28 AOIs
for i=1:size(noses,1)
    disp(['Estimation of mean heatmap values, mean local entropy, mean grayscale value, and eccentricity per area of interest (n = 28), AOI number = ' num2str(i)])
    mask_i=find(mask==i);
    for i2=1:3 % i2=1: entire 5 minutes; i2=2: first minute; i2=3: last minute
        Heatmap_m=squeeze(HMstore(i2,:,:));Heatmap_m(:);Heatmap_m=Heatmap_m(mask_i);VM(i,i2)=mean(Heatmap_m); % mean heatmap values (real)
    end
    Entropy_m=EntropyImg;Entropy_m(:);Entropy_m=Entropy_m(mask_i);VM(i,4)=mean(Entropy_m); % mean local entropy value (real)
    Grayscale_m_real=original_grayscale;Grayscale_m_real(:);Grayscale_m_real=Grayscale_m_real(mask_i);VM(i,5)=mean(Grayscale_m_real); % mean grayscale value (real)
    
    for i2=1:3 % 1=i2=1: entire 5 minutes; i2=2: first minute; i2=3: last minute
        HMstore_replica_m=squeeze(HMstore_replica(i2,:,:));HMstore_replica_m(:);HMstore_replica_m=HMstore_replica_m(mask_i);VM(i,i2+5)=mean(HMstore_replica_m);
    end
    Recreate2D_m=Recreate2D;Recreate2D_m(:);Recreate2D_m=Recreate2D_m(mask_i);VM(i,9)=mean(Recreate2D_m);
    Grayscale_m_replica=Recreate2D_replica;Grayscale_m_replica(:);Grayscale_m_replica=Grayscale_m_replica(mask_i);VM(i,10)=mean(Grayscale_m_replica); % mean grayscale value (replica)
end
VM(:,11)=Eccentricity; % eccentricity per area of interest
%% Table 4. Pearson correlation matrix among mean heatmap values, mean local entropy, mean grayscale value, and eccentricity per area of interest (n = 28).
disp('Pearson correlation matrix among mean heatmap values, mean local entropy, mean grayscale value, and eccentricity per area of interest (n = 28)')
disp(round(corr(VM(:,[1 2 3 6 7 8 4 5 10 11])),2))
%% Figure S1. Left: Scatter plot of the mean heatmap value versus eccentricity for the real painting. Right: Scatter plot of the mean grayscale value of the replica versus eccentricity. The numbers correspond to the areas of interest shown in Figure 7.
figure
subplot(1,2,1)
hold on;grid on;box on
for i=1:size(noses,1)
    scatter1=scatter(VM(i,11),VM(i,1),700,'markerfacecolor',colors(i,:),'markeredgecolor',colors(i,:));hold on;grid on % mean heatmap value versus eccentricity for the real painting
    scatter1.MarkerFaceAlpha = .3;
    text(VM(i,11),VM(i,1),num2str(i),'horizontalalignment','center')
end
set(gca,'pos',[0.05 0.09 0.42 0.89])
xlabel('Eccentricity (Real painting) (m)')
ylabel('Mean heatmap value (Real painting)')
h=findobj('FontName','Helvetica');set(h,'FontSize',20,'Fontname','Arial')

subplot(1,2,2)
hold on;grid on;box on
for i=1:size(noses,1)
    scatter1=scatter(VM(i,11),VM(i,10),700,'markerfacecolor',colors(i,:),'markeredgecolor',colors(i,:));hold on;grid on % mean grayscale value of the replica versus eccentricity
    scatter1.MarkerFaceAlpha = .4;
    text(VM(i,11),VM(i,10),num2str(i),'horizontalalignment','center')
end
set(gca,'pos',[0.57 0.09 0.42 0.89])
xlabel('Eccentricity (Real painting) (m)')
ylabel('Mean grayscale value (Replica) (0-255)')
h=findobj('FontName','Helvetica');set(h,'FontSize',20,'Fontname','Arial')
set(gcf,'color','w');
%% Figure 12. Left: Scatter plot of the mean heatmap value (real painting) versus mean local entropy value (real painting) per area of interest. Right: Scatter plot of the mean heatmap value (real painting) for the last versus first minute of viewing. The diagonal line is the line of unity. The numbers correspond to the areas of interest (AOIs) shown in Figure 7. It can be seen, for example, that AOI 16 (the face of captain Frans Banninck Cocq) received less attention in the last minute compared to the first minute.
figure;
subplot(1,2,1)
hold on;grid on;box on
for i=1:size(noses,1)
    scatter1=scatter(VM(i,4),VM(i,1),700,'markerfacecolor',colors(i,:),'markeredgecolor',colors(i,:));hold on;grid on % mean heatmap value (real painting) versus mean local entropy value (real painting)
    scatter1.MarkerFaceAlpha = .3;
    text(VM(i,4),VM(i,1),num2str(i),'horizontalalignment','center')
end
set(gca,'pos',[0.05 0.09 0.42 0.89])
xlabel('Mean entropy value (Real painting)')
ylabel('Mean heatmap value (Real painting)')
h=findobj('FontName','Helvetica');set(h,'FontSize',20,'Fontname','Arial')

subplot(1,2,2)
plot([0 4],[0 4],'k');
hold on;grid on;box on
for i=1:size(noses,1)
    scatter1=scatter(VM(i,2),VM(i,3),700,'markerfacecolor',colors(i,:),'markeredgecolor',colors(i,:));hold on;grid on % mean heatmap value (real painting) for the last versus first minute of viewing
    scatter1.MarkerFaceAlpha = .4;
    text(VM(i,2),VM(i,3),num2str(i),'horizontalalignment','center')
end
set(gca,'pos',[0.57 0.09 0.42 0.89])
xlabel('Mean heatmap value - First minute (Real painting)')
ylabel('Mean heatmap value - Last minute (Real painting)')
h=findobj('FontName','Helvetica');set(h,'FontSize',20,'Fontname','Arial')
set(gcf,'color','w');
%% Figure 14. Left: Scatter plot of mean heatmap value for the replica vs. the real painting. Right: Scatter plot of mean heatmap value for the last versus first minute of viewing for the replica Night Watch. The diagonal line is the line of unity. The numbers correspond to the areas of interest shown in Figure 7.
figure;
subplot(1,2,1)
plot([0 4],[0 4],'k');
hold on;grid on;box on
for i=1:size(noses,1)
    scatter1=scatter(VM(i,1),VM(i,6),700,'markerfacecolor',colors(i,:),'markeredgecolor',colors(i,:));hold on;grid on % mean heatmap value for the replica vs. the real painting
    scatter1.MarkerFaceAlpha = .3;
    text(VM(i,1),VM(i,6),num2str(i),'horizontalalignment','center')
end
set(gca,'pos',[0.05 0.09 0.42 0.89],'xlim',[0 6],'xtick',0:1:6)
xlabel('Mean heatmap value (Real painting)')
ylabel('Mean heatmap value (Replica)')
h=findobj('FontName','Helvetica');set(h,'FontSize',20,'Fontname','Arial')

subplot(1,2,2)
plot([0 4],[0 4],'k');
hold on;grid on;box on
for i=1:size(noses,1)
    scatter1=scatter(VM(i,7),VM(i,8),700,'markerfacecolor',colors(i,:),'markeredgecolor',colors(i,:));hold on;grid on %  mean heatmap value for the last versus first minute of viewing for the replica Night Watch
    scatter1.MarkerFaceAlpha = .4;
    text(VM(i,7),VM(i,8),num2str(i),'horizontalalignment','center')
end
set(gca,'pos',[0.57 0.09 0.42 0.89],'xlim',[0 6],'xtick',0:1:6)
xlabel('Mean heatmap value - First minute (Replica)')
ylabel('Mean heatmap value - Last minute (Replica)')
h=findobj('FontName','Helvetica');set(h,'FontSize',20,'Fontname','Arial')
set(gcf,'color','w');
%% Here some statistics are computed: (1) the recording time, (2) the last frame with data, and (3) the percentage of missing data
STATS=NaN(21,3);
for i=1:size(XA,1) % loop over 21 participants
    STATS(i,1)=(find(~isnan(XA(i,:)),1,'last')-1)/25; % recording time (300 s)
    STATS(i,2)=find(~isnan(XA(i,:)),1,'last'); % last frame
    STATS(i,3)=100*sum(isnan(XA(i,:)))/7501; % percentage of missing data
    if ismember(i,16) % participant whose recording ended earlier than 5 minutes due to an empty battery of the eye-tracker
        STATS(i,3)=100*sum(isnan(XA(i,1:STATS(i,2))))/STATS(i,2);
    end
    STATS(i,4)=sum(~isnan(XA(i,1:STATS(i,2)))); % number of frames available
end
disp(['Mean percentage eye-tracking data available = ' num2str(mean(100-(STATS(:,3)))) '% SD = ' num2str(std(100-(STATS(:,3)))) '%, min = '  num2str(min(100-(STATS(:,3)))) '%, max = ' num2str(max(100-(STATS(:,3)))) '%'])
%% Figure 11. Viewing speed of the real Night Watch, i.e., the speed of movement of the gaze point on the surface of the painting. This figure represents the median of all participants, which was subsequently filtered using a median filter with a window length of 250 (10 s).
Distance=NaN(size(XAf,1),size(XAf,2)-1);
Time=0.04:0.04:300; % Time vector (25 Hz, 7500 frames)
for i=1:21
    Distance(i,:)=25*1.895*sqrt(diff(XAf(i,:)).^2+diff(YAf(i,:)).^2)/1000; % 1 pixel is 1.895 mm, 25 Hz (width: 4535 mm (Wikipedia), 2393 pixels)
end
figure;
plot(Time,movmedian(median(Distance,'omitnan'),25*10),'k','Linewidth',3);hold on
grid on
xlabel('Elapsed time (s)')
ylabel('Viewing speed (m/s)')
set(gca,'LooseInset',[0.01 0.01 0.01 0.01],'ylim',[0 1.5],'xtick',0:10:300)
h=findobj('FontName','Helvetica'); set(h,'FontSize',24,'Fontname','Arial')
