% MATLAB script that produces all eye-tracking-related figures and analyses for the experiment with the replica Night Watch
clear all;close all;clc %#ok<CLALL>
close all
addpath('..\Functions')

% [~,mapping]=xlsread('..\mapping.xlsx','A2:G28'); % load info about the participants
original=imread('2444px-The_Night_Watch_-_HD.jpg'); % 2444 pixels horizontal, 2048 pixels vertical
original_grayscale=rgb2gray(original);

mycolormap=colormap("hsv");
ft=customcolormap([0 1],[mycolormap(52,:);0.5 0.5 0.5]);
mycolormap(1:52,:)=ft(1:5:end,:);

readraw=0; % if 1, then read Excel files with Tobii data (this will take ~30 minutes and will save a MAT file X_replica.mat)
process_reference_images=0; % if 1, then the reference images from Tobii will be mapped to the frontal image of the painting. Regression coefficients will be stored
match_to_reference_images=0; % if 1, then the Tobii image segment being gazed at is matched with the reference images. The XY coordinates are stored in CData6...mat
plot_for_participant=0; % make plot for a specific frame of a participant (useful for debugging/checking)
video_separate=0; % create video with gaze markers for a single participant on the frontal image of the painting
video_combined=0; % create video with gaze markers of all participants on the frontal image of the painting
recreate2dimage=0; % recreate frontal image from reference image and save into mat-file

window_size=[300 250 200 150 100 75 50]; % templates of different sizes cropped around the gaze coordinate

for i=1:9 % load the reference images
    fixed_rgb=imread([num2str(i) '.jpg']); % load image
    FIXED_RGB(i,:,:,:)=fixed_rgb; %#ok
    fixed=rgb2gray(fixed_rgb);
    FIXED(i,:,:)=fixed; %#ok
end
%% This block reads Tobii data and saves it in a MAT file X_replica.mat
if readraw==1 % read Excel files with Tobii data
    X=NaN(29,200000,52);
    for p=3:29 % loop over participants (first two participants do not exist)
        disp(['Read Tobii data, participant number = ' num2str(p)])
        [x,xt]=xlsread(['Participant' sprintf('%03d',p) '.xlsx'],'A2:AX300000'); %#ok<XLSRD> 
        if ismember(p,[3 4])
            x(:,51:52)=[strcmp(xt(:,1),'Accelerometer')+2*strcmp(xt(:,1),'Eye Tracker')+3*strcmp(xt(:,1),'Gyroscope') strcmp(xt(:,35),'fullstream.mp4')];
        else % files for participants 5 onward were exported with a column less, so this needs to be corrected
            x(:,6:50)=x(:,5:49);
            x(:,51:52)=[strcmp(xt(:,1),'Accelerometer')+2*strcmp(xt(:,1),'Eye Tracker')+3*strcmp(xt(:,1),'Gyroscope') strcmp(xt(:,34),'fullstream.mp4')];
        end
        X(p,1:size(x,1),1:size(x,2))=x;
    end
    lastindex=2+find(nansum(X(:,3:end,1))==0,2,'first'); %#ok<NANSUM>
    X=X(:,3:lastindex(2)-1,:); % cut off empty part of the matrix
    save('X_replica.mat','X') % save MAT file with eye-tracking data
else
    load X_replica.mat
end
%% This block reads the references images from Tobii and matches them to the large frontal image of the painting
%% Regression coefficients are saved. These regression coefficients enable
%% the mapping from the reference images to the frontal image of the painting
%% The regression coefficients are saved in a MAT file RegressionCoefficients_replica
if process_reference_images==1 % read Excel files with Tobii data
    RegressionCoefficients=NaN(9,2,7);
    for i=1:9 % loop over reference images
        disp(['Read and match the references images to the large frontal image of the painting, Reference image number = ' num2str(i)])
        points1 = detectORBFeatures(squeeze(FIXED(i,:,:)),'numlevels',14,'scalefactor',1.2); % detect ORB (i.e., corner) points from reference image
        points2 = detectORBFeatures(original_grayscale,'numlevels',14,'scalefactor',1.2); % detect ORB (i.e., corner) points from frontal image
        [features1,valid_points1] = extractFeatures(squeeze(FIXED(i,:,:)),points1); % extract feature vectors (i.e., descriptors) and their locations from reference image
        [features2,valid_points2] = extractFeatures(original_grayscale,points2); % extract feature vectors (i.e., descriptors) and their locations from frontal image
        indexPairs = matchFeatures(features1,features2); % find matching features between reference and frontal image
        matchedPoints1 = valid_points1(indexPairs(:,1),:); % matched points on reference image
        matchedPoints2 = valid_points2(indexPairs(:,2),:); % matched points on frontal image
        TobiiPoints=matchedPoints1.Location; % location of matched points on reference image
        OriginalPoints=matchedPoints2.Location; % location matched points on frontal image
        
        switch i
            case 1
                TobiiPoints_corners=[737 280]; % nose of figure at the left (Reijnier Engelen)
                OriginalPoints_corners=[150 900]; % nose of figure at the left (Reijnier Engelen)
                TobiiPoints_corners(2,:)=[916 60]; % midpoint of flag stick
                OriginalPoints_corners(2,:)=[541 473]; % midpoint of flag stick
                TobiiPoints_corners(3,:)=[1831 309]; % nose of second figure from the right, above the drummer (Rombout Kemp)
                OriginalPoints_corners(3,:)=[2230 943]; % nose of second figure from the right, above the drummer (Rombout Kemp)
                TobiiPoints_corners(4,:)=[1810 664]; % point at the bottom ring of the drum
                OriginalPoints_corners(4,:)=[2294 1600]; % point at the bottom ring of the drum
                TobiiPoints_corners(5,:)=[730 744]; % left bottom of the painting
                OriginalPoints_corners(5,:)=[57+30 1995-18]; % left bottom of the painting
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 2
                TobiiPoints_corners=[1011 309]; % nose of figure at the left (Reijnier Engelen)
                OriginalPoints_corners=[150 900]; % nose of figure at the left (Reijnier Engelen)
                TobiiPoints_corners(2,:)=[1204 93]; % midpoint of flag stick
                OriginalPoints_corners(2,:)=[541 473]; % midpoint of flag stick
                TobiiPoints_corners(3,:)=[977 765]; % left bottom of the painting
                OriginalPoints_corners(3,:)=[57+30 1995-18]; % left bottom of the painting
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 3
                TobiiPoints_corners=[775 453]; % nose of figure at the left (Reijnier Engelen)
                OriginalPoints_corners=[150 900]; % nose of figure at the left (Reijnier Engelen)
                TobiiPoints_corners(2,:)=[754 916]; % left bottom of the painting
                OriginalPoints_corners(2,:)=[57+30 1995-18]; % left bottom of the painting
                TobiiPoints_corners(3,:)=[1851 487]; % nose of second figure from the right, above the drummer (Rombout Kemp)
                OriginalPoints_corners(3,:)=[2230 943]; % nose of second figure from the right, above the drummer (Rombout Kemp)
                TobiiPoints_corners(4,:)=[1872 846]; % point at the bottom ring of the drum
                OriginalPoints_corners(4,:)=[2294 1600]; % point at the bottom ring of the drum
                TobiiPoints_corners(5,:)=[745 82]; % left top of the painting
                OriginalPoints_corners(5,:)=[62+30 53+50]; % left top of the painting
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 4
                TobiiPoints_corners=[835 546]; % nose of figure at the left (Reijnier Engelen)
                OriginalPoints_corners=[150 900]; % nose of figure at the left (Reijnier Engelen)
                TobiiPoints_corners(2,:)=[791 941]; % left bottom of the painting
                OriginalPoints_corners(2,:)=[57+30 1995-18]; % left bottom of the painting
                TobiiPoints_corners(3,:)=[1699 604]; % nose of second figure from the right, above the drummer (Rombout Kemp)
                OriginalPoints_corners(3,:)=[2230 943]; % nose of second figure from the right, above the drummer (Rombout Kemp)
                TobiiPoints_corners(4,:)=[1727 913]; % point at the bottom ring of the drum
                OriginalPoints_corners(4,:)=[2294 1600]; % point at the bottom ring of the drum
                TobiiPoints_corners(5,:)=[832 260]; % left top of the painting
                OriginalPoints_corners(5,:)=[62+30 53+50]; % left top of the painting
                TobiiPoints_corners(6,:)=[1779 232]; % right top of the painting
                OriginalPoints_corners(6,:)=[2419-60 53+50]; % right top of the painting
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 5
                TobiiPoints_corners=[589 1029]; % left bottom of the painting
                OriginalPoints_corners=[57+30 1995-18]; % left bottom of the painting
                TobiiPoints_corners(2,:)=[569 163]; % left top of the painting
                OriginalPoints_corners(2,:)=[62+30 53+50]; % left top of the painting
                TobiiPoints_corners(3,:)=[1655 85]; % right top of the painting
                OriginalPoints_corners(3,:)=[2419-60 53+50]; % right top of the painting
                TobiiPoints_corners(4,:)=[1731 1042]; % right bottom of the painting
                OriginalPoints_corners(4,:)=[2419-60 1995-18]; % right bottom of the painting
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 6
                TobiiPoints_corners=[417 963]; % left bottom of the painting
                OriginalPoints_corners=[57+30 1995-18]; % left bottom of the painting
                TobiiPoints_corners(2,:)=[414 134]; % left top of the painting
                OriginalPoints_corners(2,:)=[62+30 53+50]; % left top of the painting
                TobiiPoints_corners(3,:)=[1408 127]; % right top of the painting
                OriginalPoints_corners(3,:)=[2419-60 53+50]; % right top of the painting
                TobiiPoints_corners(4,:)=[1424 954]; % right bottom of the painting
                OriginalPoints_corners(4,:)=[2419-60 1995-18];% right bottom of the painting
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 7
                TobiiPoints_corners=[63 407]; % nose of figure at the left (Reijnier Engelen)
                OriginalPoints_corners=[150 900]; % nose of figure at the left (Reijnier Engelen)
                TobiiPoints_corners(2,:)=[1165 114]; % right top of the painting
                OriginalPoints_corners(2,:)=[2419-60 53+50]; % right top of the painting
                TobiiPoints_corners(3,:)=[1138 914]; % right bottom of the painting
                OriginalPoints_corners(3,:)=[2419-60 1995-18]; % right bottom of the painting
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 8
                TobiiPoints_corners=[1042 227]; % tip of pike at the right (held by Walich Schellingwou)
                OriginalPoints_corners=[2268 643]; % tip of pike at the right (held by Walich Schellingwou)
                TobiiPoints_corners(2,:)=[1099 824]; % right bottom of the painting
                OriginalPoints_corners(2,:)=[2419-60 1995-18]; % right bottom of the painting
                TobiiPoints_corners(3,:)=[82 95]; % midpoint of flag stick
                OriginalPoints_corners(3,:)=[541 473]; % midpoint of flag stick
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
            case 9
                TobiiPoints_corners=[176 1067]; % left bottom of the painting
                OriginalPoints_corners=[57+30 1995-18]; % left bottom of the painting
                TobiiPoints_corners(2,:)=[27 58]; % left top of the painting
                OriginalPoints_corners(2,:)=[62+30 53+50]; % left top of the painting
                TobiiPoints_corners(3,:)=[1232 66]; % right top of the painting
                OriginalPoints_corners(3,:)=[2419-60 53+50]; % right top of the painting
                TobiiPoints_corners(4,:)=[1269 899]; % right bottom of the painting
                OriginalPoints_corners(4,:)=[2419-60 1995-18]; % right bottom of the painting
                TobiiPoints_corners=repmat(TobiiPoints_corners,30,1);
                OriginalPoints_corners=repmat(OriginalPoints_corners,30,1);
        end
        TobiiPoints=[TobiiPoints;TobiiPoints_corners]; %#ok<AGROW>
        OriginalPoints=[OriginalPoints;OriginalPoints_corners]; %#ok<AGROW>
        
        Offsetvector=-6000:50:6000; % vector of possible values for b6 and b7
        CC=NaN(length(Offsetvector),length(Offsetvector),2);
        OriginalPoints_pred=NaN(length(Offsetvector),length(Offsetvector),size(TobiiPoints,1),2);
        bx=NaN(length(Offsetvector),length(Offsetvector),5);
        by=NaN(length(Offsetvector),length(Offsetvector),5);
        
        for xxi=1:length(Offsetvector) % loop over b6
            for yyi=1:length(Offsetvector) % loop over b7
                o1=TobiiPoints(:,2)-Offsetvector(yyi); % y-b7
                o2=TobiiPoints(:,1)-Offsetvector(xxi); % x-b6
                TobiiPoints=[TobiiPoints(:,1:2) o2./bound(o1,-150,150) o1./bound(o2,-150,150)]; % x, y, (x-b6)/(y-b7), (y-b7)/(x-b6)
                warning off all
                bx(xxi,yyi,:)=robustfit(TobiiPoints,OriginalPoints(:,1)); % perform regression analysis to predict x on the frontal image (Eq. 1)
                by(xxi,yyi,:)=robustfit(TobiiPoints,OriginalPoints(:,2)); % perform regression analysis to predict y on the frontal image (Eq. 2)
                warning on all
                OriginalPoints_pred(xxi,yyi,:,1)=bx(xxi,yyi,1)+bx(xxi,yyi,2)*TobiiPoints(:,1)+bx(xxi,yyi,3)*TobiiPoints(:,2)+bx(xxi,yyi,4)*TobiiPoints(:,3)+bx(xxi,yyi,5)*TobiiPoints(:,4); % predicted x on the frontal image (Eq. 1)
                OriginalPoints_pred(xxi,yyi,:,2)=by(xxi,yyi,1)+by(xxi,yyi,2)*TobiiPoints(:,1)+by(xxi,yyi,3)*TobiiPoints(:,2)+by(xxi,yyi,4)*TobiiPoints(:,3)+by(xxi,yyi,5)*TobiiPoints(:,4); % predicted y on the frontal image (Eq. 2)
                CC(xxi,yyi,:)=diag(corr(squeeze(OriginalPoints_pred(xxi,yyi,:,:)),OriginalPoints)); % correlation coefficient between points on the reference image and the frontal image
            end
        end
        [ovx_x,ovy_x]=find(CC(:,:,1)==max(max(CC(:,:,1))),1,'first'); % obtain indexes of maximum correlation coefficient between x-coordinates on the reference image and the frontal image
        [ovx_y,ovy_y]=find(CC(:,:,2)==max(max(CC(:,:,2))),1,'first'); % obtain indexes of maximum correlation coefficient between y-coordinates on the reference image and the frontal image
        
        RegressionCoefficients(i,1,:)=[squeeze(bx(ovx_x,ovy_x,:))' Offsetvector(ovx_x) Offsetvector(ovy_x)]; % store the 7 regression coefficients (bx1-7) that correspond to the maximum correlation
        RegressionCoefficients(i,2,:)=[squeeze(by(ovx_y,ovy_y,:))' Offsetvector(ovx_y) Offsetvector(ovy_y)]; % store the 7 regression coefficients (by1-7) that correspond to the maximum correlation
    end % end of reference images loop
    save('RegressionCoefficients_replica.mat','RegressionCoefficients'); % 9 reference images x 2 coordinates (x,y) x 7 regression coefficients
else
    load RegressionCoefficients_replica.mat
end
%% Figure S2. Bottom: Reconstructed frontal view from a reference image of the replica Night Watch. The reconstruction was performed using Eqs. (1) and (2).
if recreate2dimage==1
    clear OriginalPoints_pred_All OriginalPoints_pred
    Recreate2D=NaN(size(original_grayscale,1),size(original_grayscale,2));
    for i=6 % use reference image 6 only
        bx=squeeze(RegressionCoefficients(i,1,:))'; % seven regression coefficients for reference image and x-coordinate (Eq. 1)
        by=squeeze(RegressionCoefficients(i,2,:))'; % seven regression coefficients for reference image and y-coordinate (Eq. 2)
        for II=1:1080 % loop over y (vertical) coordinates of reference image
            for JJ=1:1920 % loop over x (horizontal) coordinates of reference image
                
                o1=II-bx(7); % y-b7
                o2=JJ-bx(6); % x-b6
                TobiiPoints=[JJ II o2/bound(o1,-150,150) o1/bound(o2,-150,150)]; % x, y, (x-b6)/(y-b7), (y-b7)/(x-b6)
                OriginalPoints_pred(1)=bx(1)+bx(2)*TobiiPoints(1)+bx(3)*TobiiPoints(2)+bx(4)*TobiiPoints(3)+bx(5)*TobiiPoints(4); % predicted x on the frontal image (Eq. 1)
                
                o1=II-by(7); % y-b7
                o2=JJ-by(6); % y-b7
                TobiiPoints=[JJ II o2/bound(o1,-150,150) o1/bound(o2,-150,150)]; % x, y, (x-b6)/(y-b7), (y-b7)/(x-b6)
                OriginalPoints_pred(2)=by(1)+by(2)*TobiiPoints(1)+by(3)*TobiiPoints(2)+by(4)*TobiiPoints(3)+by(5)*TobiiPoints(4); % predicted y on the frontal image (Eq. 2)
                
                if OriginalPoints_pred(2) > 1 && OriginalPoints_pred(2) <= size(original_grayscale,1) && OriginalPoints_pred(1) > 1 && OriginalPoints_pred(1) <= size(original_grayscale,2) && isnan(Recreate2D(round(OriginalPoints_pred(2)),round(OriginalPoints_pred(1)))) % only if within points and pixel is not predicted yet
                    Recreate2D(round(OriginalPoints_pred(2)),round(OriginalPoints_pred(1)))=FIXED(i,II,JJ);
                end
            end % end of x (horizontal) coordinates loop
        end % end of y (vertical) coordinates loop
    end % end of reference images loop
    
    [II,JJ]=find(isnan(Recreate2D)); % find pixels that were not predicted
    
    for i=1:length(II) % loop over pixels that were not predicted, to fill with neighboring values
        if II(i)>1&&JJ(i)>1&&II(i)<size(Recreate2D,1)&&JJ(i)<size(Recreate2D,2)
            pixs=Recreate2D(II(i)-1:II(i)+1,JJ(i)-1:JJ(i)+1);
            Recreate2D(II(i),JJ(i))=mean(pixs(:), 'omitnan');
        end
    end
    Recreate2D_replica=Recreate2D; % save frontal image of the painting recreated from reference images from the replica Night Watch experiment
    save('Recreate2D_replica.mat','Recreate2D_replica')
end
%% This block matches the Tobii image segment being gazed at with the images
%% A loop is run over all participants and video frames
%% The results are outputted as x and y coordinates of the matched reference image (participant x frame x reference image x window size)
%% The results are saved in CData6... .mat
if match_to_reference_images==1
    [MaxCor,II,JJ]=deal(NaN(29,15000,9,length(window_size)));
    Coordinates_store=NaN(29,15000,2);
    FIXED_GPU=gpuArray(FIXED);
    
    close all
    for p=1:27 % loop over participants (corresponding to participant 3-29)
        Xp=squeeze(X(p+2,:,:)); % select data for participant p
        Xp=Xp(Xp(:,52)==1,:); % only select data points for which video recording is available
        
        videoname=mapping{p,3};
        vidObj=VideoReader(['..\..\Experiment - Replica TU\Project1\Media\' videoname '.mp4']);
        
        Time=Xp(:,1);
        Time=Time/10^6-Time(end)/10^6+vidObj.Duration; % create time vector that matches the video time
        
        eyetracker=find(Xp(:,51)==2 & Time<=max(Time)-0.5); % find indexes of eye tracker data and discard last 0.5 seconds
        
        Xp=Xp(eyetracker,:); % select only the eye tracker data
        Gaze=Xp(:,16:17); % gaze point x, gaze point y (in pixels)
        Time=Time(eyetracker,:); % time vector
        FrameNumber=ceil(Time*25.01762); % compute video frame number that corresponds to gaze data
        
        % compute miss_start and miss_end; they define the gaps in the data, for example due to blinks
        miss=isnan(Gaze(:,1)); % missing eye tracking data
        miss_start=1+find(diff(miss)>0); % start indexes of intervals with NaNs
        miss_end=NaN(size(miss_start));
        for i=1:length(miss_start)
            try
                miss_end(i)=miss_start(i)+find(Gaze(miss_start(i):end,1)>0,1,'first')-2; % end indexes of intervals with NaNs
            catch error
            end
        end
        
        GazeC=Gaze;
        for i=1:length(miss_start) % increase the gap sizes with 2 frames. This is done because the eyelid opening and closing take about 2 frames each. These data need to be interpolated
            try
                GazeC(miss_start(i)-2:miss_end(i)+2)=NaN;
            catch error
            end
        end
        GazeCF=NaN(size(GazeC));
        GazeCF(:,1)=fixgaps(GazeC(:,1)); % linearly interpolate eye gaze gaps
        GazeCF(:,2)=fixgaps(GazeC(:,2)); % linearly interpolate eye gaze gaps
        GazeCF=movmedian(GazeCF,20); % median filter to cancel out noise
        
        for i=1:length(miss_start)
            if miss_end(i)-miss_start(i)>50 % if the gap is greater than 0.5 s
                GazeCF(miss_start(i):miss_end(i),:)=NaN; % then set the data to NaN. We do not want long gaps (> 0.5 s) to be interpolated as this would be misleading (suggesting there is data while there is none)
            end
        end
        
        vr = VideoWriter(['Moving_with_marker_participant_no_frameno_' num2str(p)], 'Mpeg-4'); %#ok<*TNMLP>
        vr.Quality=100;vr.FrameRate=25;
        open(vr);
        begin=1; % can be changed if wanting to show a specific frame
        
        for f=begin:max(FrameNumber) % loop over frames to be analysed
            disp(['Coordinates of the match of the Tobii image segment being gazed at with the reference images,  Participant number = ' num2str(p) ', Frame number = ' num2str(f)])
            
            if rem(f,100)==1 || f==begin % loading batches of 100 frames, to speed up the code
                fend=f+99;
                if fend>vidObj.NumFrames
                    fend=vidObj.NumFrames;
                end
                moving = read(vidObj, [f fend]); % read batches of the video, e.g. 1-100, 101-200, etc
                fbegin=f;
            end
            moving_with_marker=moving(:,:,:,f-fbegin+1);
            moving_gray=rgb2gray(squeeze(moving(:,:,:,f-fbegin+1))); % 1080 x 1920 matrix
            eye_frame=find(FrameNumber==f); % eye-gaze indexes of frame
            if ~isempty(eye_frame) && ~isnan(nanmean(GazeCF(eye_frame,1))) %#ok<NANMEAN> % if there is data for this frame f
                GazeFrame=nanmean(GazeCF(eye_frame,:),1); %#ok<NANMEAN> % gaze x and y corresponding to frame f
                x=round(GazeFrame(:,1)); % x gaze coordinate in Tobii image; round to nearest integer (pixel)
                y=round(GazeFrame(:,2)); % y gaze coordinate in Tobii image; round to nearest integer (pixel)
                moving_with_marker = insertShape(moving_with_marker,'FilledCircle',[x y 15],'color',[0 255 0],'opacity',1); % add circular green marker to the Tobii video
                Coordinates_store(p,f,:)=[x y]; % x, y gaze coordinates
                
                for ws=1:length(window_size)
                    
                    AOI=gpuArray(imcrop(moving_gray,[x-window_size(ws) y-window_size(ws) 2*window_size(ws) 2*window_size(ws)])); % rectangle around gaze point (e.g., if window_size(ws)=300, then AOI is 601 x 601 pixels)
                    for i=1:9 % loop over reference images
                        if std(std(double(AOI)))>0 % there needs to be at least some variation in the AOI (not only white pixels by looking at the reflection)
                            
                            C=normxcorr2(AOI, squeeze(FIXED_GPU(i,:,:))); % normalized two-dimensional cross-correlation
                            MaxCor(p,f,i,ws)=max(C(:)); % find maximum correlation between AOI and reference image
                            [II(p,f,i,ws),JJ(p,f,i,ws)]=find(C==MaxCor(p,f,i,ws),1,'first'); % II: y coordinate, JJ: x coordinate
                            II(p,f,i,ws)=II(p,f,i,ws)-window_size(ws); % compute the y gaze coordinate in the reference image of maximal correlation
                            JJ(p,f,i,ws)=JJ(p,f,i,ws)-window_size(ws); % compute the x gaze coordinate in the reference image of maximal correlation
                            
                            if y>1080-window_size(ws) % if the participant gazed at the bottom of the Tobii image, then apply correction to account for cropping effects
                                II(p,f,i,ws)=II(p,f,i,ws)+y-1080+window_size(ws);
                            end
                            if x>1920-window_size(ws) % if the participant gazed at the right of the Tobii image, then apply correction to account for cropping effectss
                                JJ(p,f,i,ws)=JJ(p,f,i,ws)+x-1920+window_size(ws);
                            end
                            
                            % Example verification
                            % participant 1, frame 202, window_size(ws) = 300
                            % Tobii gaze coordinates 272, 1630 (y, x)
                            % so the participant gazed at the upper right, which is why the AOI is cropped at the top and at the right
                            % size(AOI): 572 (1:572), 591 (1330:1920) (y, x)
                            
                            % the size of C (output of normmxcor2) = size(moving_gray)+size(AOI)-1
                            % coordinate of maximum correlation according to normxcorr2, including padding: i = 1080, j = 2354 (y, x)
                            % so the left top of the match is at 508, 1763 in the matrix excluding padding (y, x)
                            % Thus, the coordinates of the gaze in the reference image are
                            % 779, 2063 (y, x)
                            
                            % AOI size subtracted (per https://nl.mathworks.com/help/images/ref/normxcorr2.html)
                            % yoffSet = ypeak-size(onion,1);xoffSet = xpeak-size(onion,2); to account for padding of normxcorr2
                            % i-size(AOI,1) = 508; j-size(AOI,2) = 1763 (y, x; so this is the left top of the match)
                            
                        end
                    end % end of reference images loop
                end % end of window sizes loop
            end
            
            R=moving_with_marker;
            writeVideo(vr,R);
        end
        close(vr)
        filename=['Cdata_replica_' num2str(round(now*36*2400)) '.mat'];
        save(filename,'MaxCor','II','JJ','Coordinates_store')
    end % end of participants loop
else
    load Cdata_replica_63812886879.mat
end
%% This block computes the points on the frontal image of the painting with the help of the regression coefficients
%% The XY coordinates are stored in a matrix 'OriginalPoints_pred_All' (participants x frames x reference image no x window size)
clear OriginalPoints_pred_All OriginalPoints_pred
OriginalPoints_pred_All=NaN(size(II,1)-2,size(II,2),size(II,3),size(II,4),2);
for p=1:27 % loop over participants
    disp(['Compute points on original painting based on regression coefficients, Participant number = ' num2str(p)])
    for i=1:9 % loop over reference images
        bx=squeeze(RegressionCoefficients(i,1,:))';
        by=squeeze(RegressionCoefficients(i,2,:))';
        for i2=1:7 % loop over AOI window sizes
            for s=1:size(II,2) % loop over frames
                if II(p,s,i,i2) >= 1 && JJ(p,s,i,i2) >= 1 && II(p,s,i,i2) <= 1080 && JJ(p,s,i,i2) <= 1920 % only make the prediction within bounds
                    
                    o1=II(p,s,i,i2)-bx(7); % y-b7
                    o2=JJ(p,s,i,i2)-bx(6); % x-b6
                    TobiiPoints=[JJ(p,s,i,i2) II(p,s,i,i2) o2/bound(o1,-150,150) o1/bound(o2,-150,150)]; % x, y, (x-b6)/(y-b7), (y-b7)/(x-b6)
                    OriginalPoints_pred(1)=bx(1)+bx(2)*TobiiPoints(1)+bx(3)*TobiiPoints(2)+bx(4)*TobiiPoints(3)+bx(5)*TobiiPoints(4); % predicted x on the frontal image (Eq. 1)
                    
                    o1=II(p,s,i,i2)-by(7);
                    o2=JJ(p,s,i,i2)-by(6);
                    TobiiPoints=[JJ(p,s,i,i2) II(p,s,i,i2) o2/bound(o1,-150,150) o1/bound(o2,-150,150)]; % x, y, (x-b6)/(y-b7), (y-b7)/(x-b6)
                    OriginalPoints_pred(2)=by(1)+by(2)*TobiiPoints(1)+by(3)*TobiiPoints(2)+by(4)*TobiiPoints(3)+by(5)*TobiiPoints(4); % predicted y on the frontal image (Eq. 2)
                    
                    OriginalPoints_pred_All(p,s,i,i2,1:2)=OriginalPoints_pred; % predict x and y
                end
            end
        end
    end
end
%% Compute stability scores for each reference image
%% The stability score is computed over a sliding window for the different window sizes
%% A high stability score is obtained when distances are relatively
%% constant, i.e., when the same location on the frontal image of the painting is identified
%% The stability score is helpful for determining the best match among window sizes and reference images. It compares coordinate estimates of subsequent window sizes for a given reference image
[stability_score,stability_score_index]=deal(NaN(size(II,1),size(II,2),size(II,3)));
wl=2; % sliding window length
for p=1:27 % loop over participants
    disp(['Compute stability scores for each reference image, Participant number = ' num2str(p)])
    for f=1:size(OriginalPoints_pred_All,2) % loop over frames
        for i=1:9 % loop over reference images
            OriginalPoints_pred_Sample_x=squeeze(OriginalPoints_pred_All(p,f,i,:,1));
            OriginalPoints_pred_Sample_y=squeeze(OriginalPoints_pred_All(p,f,i,:,2));
            
            ssd=NaN(size(II,4)-1,1);
            for i2=1:size(II,4)-1 % loop over window sizes
                ssd(i2)=pdist([OriginalPoints_pred_Sample_x(i2:i2+1) OriginalPoints_pred_Sample_y(i2:i2+1)]); % subsequent distance between predicted coordinates on the frontal image
            end
            ssd2=NaN(size(II,4)-wl,1);
            for i2=1:length(ssd)-wl+1
                ssd2(i2)=mean(ssd(i2:i2+wl-1)); % mean stability over wl points (lower means more consistent prediction)
            end
            [stability_score(p,f,i),stability_score_index(p,f,i)]=min(ssd2); % start index of stability interval for reference image i
        end
    end
end
%% Determining the final coordinate on the frontal image of the painting
%% From all candidate matching points on the frontal image identified above, the most appropriate one is selected based on stability and correlation coefficients
%% The result is XX and YY, i.e., the actual gaze point on the frontal image of the painting
[XX,YY]=deal(NaN(size(II,1),size(II,2)));
SelectedPoints=deal(NaN(size(II,1),size(II,2),2));
for p=1:27 % loop over participants
    disp(['Select the most appropriate point on the frontal image of the painting, Participant number = ' num2str(p)])
    for f=1:size(II,2) % loop over frames
        x=squeeze(OriginalPoints_pred_All(p,f,:,:,1));
        y=squeeze(OriginalPoints_pred_All(p,f,:,:,2));
        
        MaxCori=NaN(size(MaxCor,3),1);
        for i=1:size(MaxCor,3) % loop over reference images
            MaxCori(i)=squeeze(MaxCor(p,f,i,squeeze(stability_score_index(p,f,i)))); % look up the correlation coefficient of the template match that corresponds to the template size that yielded the best stability
        end
        ranks=-MaxCori+squeeze(stability_score(p,f,:))/100; % combine correlation coefficient and stability score using a heuristic weighting (lower score is better
        [~,b]=min(ranks); % reference image number of best match between the template and the reference image
        b2=stability_score_index(p,f,b); % corresponding template size index
        XX(p,f)=median(x(b,b2:b2+wl)); % x-coordinate of gaze point on the frontal image of the painting
        YY(p,f)=median(y(b,b2:b2+wl)); % y-coordinate of gaze point on the frontal image of the painting
        SelectedPoints(p,f,:)=[b b2]; % reference image number and template size number of best match
    end
end

if plot_for_participant==1
    %% This block shows a selected frame of a selected participant
    %% This block does not have a fundamental role in creating the results but can be used for checking the algorithm
    close all
    
    p=1;f=3000;
    
    videoname=mapping{p,3};
    x=squeeze(OriginalPoints_pred_All(p,f,:,:,1));
    disp('Example of a selected frame of a selected participant')
    disp('Coordinate x')
    disp(round(x))
    
    y=squeeze(OriginalPoints_pred_All(p,f,:,:,2));
    disp('Coordinate y')
    disp(round(y))
    disp(round(y))
    disp('Maxcor')
    disp(squeeze(MaxCor(p,f,:,:)))
    disp('Maxcor, stability score, ranks, stability score index')
    disp([median(squeeze(MaxCor(p,f,:,stability_score_index(p,f,:))),2,'omitnan') squeeze(stability_score(p,f,:)) ranks squeeze(stability_score_index(p,f,:)) ])
    disp('Selected coordinates')
    disp([XX(p,f) YY(p,f)])
    disp('Selected points')
    disp(squeeze(SelectedPoints(p,f,:)))
    
    vidObj=VideoReader(['..\..\Experiment - Replica TU\Project1\Media\' videoname '.mp4']);
    
    moving = read(vidObj, f); % read batches of the video, e.g. 1-100, 101-200, etc
    moving_with_marker = insertShape(moving,'FilledCircle',[Coordinates_store(p,f,1) Coordinates_store(p,f,2) 15],'color',[0 255 0],'opacity',1); % add circular green marker to the Tobii video
    
    figure; % show Tobii frame with marker
    imshow(moving_with_marker)
    axis equal
    
    set(gca,'LooseInset',[0.01 0.01 0.01 0.01])
    
    original_with_marker=original;
    for i=1:size(x,1)
        for j=1:size(x,2)
            if ~isnan(x(i,j))&&~isnan(y(i,j))
                original_with_marker = insertShape(original_with_marker,'FilledCircle',[x(i,j) y(i,j) 15],'color',255*[1 1 1],'opacity',1); % add circular green marker to the Tobii video
                original_with_marker = insertText(original_with_marker,[x(i,j) y(i,j)],[num2str(round(x(i,j))) ', ' num2str(round(y(i,j)))]); % add circular green marker to the Tobii video
            end
        end
    end
    original_with_marker = insertShape(original_with_marker,'FilledCircle',[XX(p,f) YY(p,f) 10],'color','red','opacity',1); % add circular green marker to the Tobii video
    
    figure; % show original picture with markers
    imshow(original_with_marker)
end
%% Finally, after processing all frames, a median filter with a window length of 5 (0.2 s) is applied to the final x and y coordinates to filter out potential outliers
% Also, out-of-bounds matches are set to NaN
XXf=medfilt1(XX',5,'omitnan')';
YYf=medfilt1(YY',5,'omitnan')';
XXf(XXf<0-100)=NaN;
XXf(XXf>2444+100)=NaN;
YYf(YYf<0-100)=NaN;
YYf(YYf>2048+100)=NaN;
startindex=[256 72 1 592 465 512 489 212 218 199 510 598 483 450 427 470 485 646 372 218 42 347 732 508 120 424 580]; % frame numbers per participant when the eye-gaze first lands on the painting
%% Create videos
%% 1. Create video with gaze markers for a single participant on the frontal image of the painting
if video_separate==1
    original_black=zeros(size(original,2),size(original,1));
    for p=15
        vr = VideoWriter(['MappingR2_' num2str(p)], 'Mpeg-4'); %#ok<*TNMLP>
        vr.Quality=100;vr.FrameRate=25;
        open(vr);
        
        videoname=mapping{p,3};
        vidObj=VideoReader(['..\Project1\Media\' videoname '.mp4']);
        
        for f=startindex(p):startindex(p)+7500
            disp(['Mapping illustration for all participants separately, Frame number = ' num2str(f)])
            moving = read(vidObj, f); % read batches of the video, e.g. 1-100, 101-200, etc
            if ~isnan(Coordinates_store(p,f,1))&&~isnan(Coordinates_store(p,f,2))
                moving_with_marker = insertShape(moving,'FilledCircle',[Coordinates_store(p,f,1) Coordinates_store(p,f,2) 25],'color',[0 255 0],'opacity',1); % add circular green marker to the Tobii video
            end
            clear R
            R(151:150+size(moving_with_marker,1),:,:)=moving_with_marker;
            R=insertText(R,[10 10],['Frame nummer = ' num2str(f-startindex(p)+1)],'fontsize',30,'BoxColor','black','BoxOpacity',0,'TextColor','green');
            R=insertText(R,[10 50],['Tijd = ' sprintf('%0.2f',(f-startindex(p))/25) ' s'],'fontsize',30,'BoxColor','black','BoxOpacity',0,'TextColor','green');
            R=insertText(R,[620 1],'Camera op hoofd','fontsize',90,'BoxColor','black','BoxOpacity',0,'TextColor','green');
            original_with_marker=original;
            if ~isnan(XXf(p,f))&&~isnan(YYf(p,f))
                original_with_marker= insertShape(original,'FilledCircle',[XXf(p,f) YYf(p,f) 30],'color',[0 255 0],'opacity',1); % add circular green marker to the Tobii video
            end
            original_with_marker=imresize(original_with_marker,0.7);
            R(126:125+size(original_with_marker,1),2001:2000+size(original_with_marker,2),:)=original_with_marker;
            R=insertText(R,[2450 1],'Photo Night Watch','fontsize',90,'BoxColor','black','BoxOpacity',0,'TextColor','green');
            %  figure;imshow(R);pause
            writeVideo(vr,R);
            
        end
        
        close(vr);
    end
end
%% 2. Create video with gaze markers of all participants on the frontal image of the painting
if video_combined==1
    vr = VideoWriter('Meta-video_replica', 'Mpeg-4'); %#ok<*TNMLP>
    vr.Quality=100;vr.FrameRate=25;
    open(vr);
    
    Rcircle=15;
    original_black=zeros(size(original,2),size(original,1));
    original_with_marker = insertShape(original_black,'FilledCircle',[50,50,Rcircle],'color',[0 255 0],'opacity',1); % create circular green marker
    Circle=double(original_with_marker(50-Rcircle:50+Rcircle,50-Rcircle:50+Rcircle,2));
    Circle=uint8(Circle/1.5);
    
    for f=1:7500 % 5 minutes of data
        original_with_marker=original;
        disp(['Video creation with markers of all persons, Frame number = ' num2str(f)])
        for p=1:27
            fillx=round(XXf(p,f+startindex(p)-1)-Rcircle:XXf(p,f+startindex(p)-1)+Rcircle);
            filly=round(YYf(p,f+startindex(p)-1)-Rcircle:YYf(p,f+startindex(p)-1)+Rcircle);
            if isempty(find(~isnan(fillx)==0,1))&&isempty(find(fillx<=0,1))&&isempty(find(fillx>2444,1))&&isempty(find(~isnan(filly)==0,1))&&isempty(find(filly<=0,1))&&isempty(find(filly>2048,1))
                original_with_marker(filly,fillx,2) = original_with_marker(filly,fillx,2)+Circle;
            end
        end
        writeVideo(vr,original_with_marker);
    end
    
    close(vr);
end
%% This block creates the heatmap. It does so by adding a patch for each participant and frame
% Figure 13. Heatmap of the 27 participants combined for a replica of The Night Watch. The heatmap was normalized by dividing by the number of patches and multiplying by 1000.

close all
clear HMstore
Rcircle=30;
original_with_marker = insertShape(zeros(500,500,3),'FilledCircle',[250,250,Rcircle],'color',[0 255 0],'opacity',1); % add circular green marker to black background
Circle=double(original_with_marker(250-Rcircle:250+Rcircle,250-Rcircle:250+Rcircle,2)); % cut out the circle
Circle=Circle/255; % we are left with a circle of radius Rcircle

HMstore=NaN(84,2048,2444); % 27 participants x 3 + 3
for hm=1:84 % create 3 heatmaps and heatmaps for individual participants
    switch hm
        case 1
            IDs=1:27; % all participants
            V=1:7500; % entire 5 minutes
        case 2
            IDs=1:27; % all participants
            V=1:1500; % first minute
        case 3
            IDs=1:27; % all participants
            V=6001:7500; % last minute
    end
    
    if ismember(hm,4:30)
        IDs=hm-3; % create heatmaps for individual participants
        V=1:7500; % entire 5 minutes
    end
    if ismember(hm,31:57)
        IDs=hm-30; % create heatmaps for individual participants
        V=1:1500; % first minute
    end
    if ismember(hm,58:84)
        IDs=hm-57; % create heatmaps for individual participants
        V=6001:7500; % last minute
    end
    
    original_black=zeros(size(original,2),size(original,1));
    
    if hm==1 % if first heatmap, with all participants
        [XA,YA,XAf,YAf]=deal(NaN(27,7501));
    end
    counter=0;
    for i=1:length(IDs) % loop over all included participants
        p=IDs(i);
        disp(['Heatmap creation, Participant number = ' num2str(p)])
        if hm==1
            XA(p,:)=XX(p,startindex(p):startindex(p)+7500); % store 5 minutes of data since the starting moment of the trial
            YA(p,:)=YY(p,startindex(p):startindex(p)+7500);
            XAf(p,:)=XXf(p,startindex(p):startindex(p)+7500);
            YAf(p,:)=YYf(p,startindex(p):startindex(p)+7500);
        end
        for f=V % loop over frames
            fillx=(XXf(p,f+startindex(p)-1)-Rcircle:XXf(p,f+startindex(p)-1)+Rcircle); % create x coordinate range for frame f
            filly=(YYf(p,f+startindex(p)-1)-Rcircle:YYf(p,f+startindex(p)-1)+Rcircle); % create y coordinate range for frame f
            fillxr=round(fillx); % x gaze coordinate in Tobii image; round to nearest integer (pixel)
            fillyr=round(filly); % y gaze coordinate in Tobii image; round to nearest integer (pixel)
            if isempty(find(~isnan(fillxr)==0,1))&&isempty(find(fillxr<=0,1))&&isempty(find(fillxr>2444,1))&&isempty(find(~isnan(fillyr)==0,1))&&isempty(find(fillyr<=0,1))&&isempty(find(fillyr>2048,1))
                original_black(fillxr,fillyr)=original_black(fillxr,fillyr)+Circle; % add the circular patch for each frame (5 minutes of data since the starting moment) and each participant
                counter=counter+1;
            end
        end
    end
    HMstore(hm,:,:)=1000*original_black'./counter; % heatmap values
    
    % Here we output the heatmap
    if ismember(hm,1:3)
        opengl software
        original_grayscale3(:,:,1)=original_grayscale;
        original_grayscale3(:,:,2)=original_grayscale;
        original_grayscale3(:,:,3)=original_grayscale;
        figure;
        imshow(original_grayscale3);hold on % frontal image of the painting
        OverlayImage = imshow(1000*original_black'./counter);
        set(OverlayImage, 'AlphaData',0.6)
        
        colormap(mycolormap);
        disp(['Maximum heatmap value: ' num2str(max(max(1000*original_black./counter)))])
        colorbar
        axis equal
        if ismember(hm,1:3)
            caxis([0 10]);
        else
            caxis([0 18]);
        end
        set(gca,'TickDir','out','LooseInset',[0.01 0.01 0.01 0.01])
        h=findobj('FontName','Helvetica'); set(h,'FontSize',24,'Fontname','Arial');
        set(gca,'TickDir','out','LooseInset',[0.01 0.01 0.01 0.01])
        set(gca,'Position',[0.00 0.03 0.94 0.94])
    end
end
HMstore_replica=HMstore;
save('HMstore_replica.mat','HMstore_replica','-v7.3')
%% Here some statistics are computed: (1) the recording time, (2) the last frame with data, and (3) the percentage of missing data
STATS=NaN(21,3);
for i=1:size(XA,1) % loop over 27 participants
    STATS(i,1)=(find(~isnan(XA(i,:)),1,'last')-1)/25; % recording time (300 s)
    STATS(i,2)=find(~isnan(XA(i,:)),1,'last'); % last frame
    STATS(i,3)=100*sum(isnan(XA(i,:)))/7501; % percent of missing data
    STATS(i,4)=sum(~isnan(XA(i,1:STATS(i,2)))); % number of frames available
end
disp(['Mean percentage eye-tracking data available = ' num2str(mean(100-(STATS(:,3)))) '% SD = ' num2str(std(100-(STATS(:,3)))) '%, min = '  num2str(min(100-(STATS(:,3)))) '%, max = ' num2str(max(100-(STATS(:,3)))) '%'])
%% Figure S3. Viewing speed of the replica Night Watch, i.e., the speed of movement of the gaze point on the surface of the painting.
Distance=NaN(size(XAf,1),size(XAf,2)-1);
Time=0.04:0.04:300; % Time vector (25 Hz, 7500 frames)
for i=1:27
    Distance(i,:)=25*0.738*sqrt(diff(XAf(i,:)).^2+diff(YAf(i,:)).^2)/1000; % 1 pixel is 0.738 mm, 25 Hz (width: 1695 mm, 2297 pixels)
end
figure;
plot(Time,movmedian(median(Distance,'omitnan'),25*10),'k','Linewidth',3);hold on
grid on
xlabel('Elapsed time (s)')
ylabel('Viewing speed (m/s)')
set(gca,'LooseInset',[0.01 0.01 0.01 0.01],'ylim',[0 .35],'xtick',0:10:300)
h=findobj('FontName','Helvetica'); set(h,'FontSize',24,'Fontname','Arial')

